#include <map>
#include <string>

#define NUM_BOND  167

typedef struct {
       char label[10];
       char first[5];
       char second[5];
       float lower_limit;
       float upper_limit;
       int   type;
} _ALLOWED_BOND;

static _ALLOWED_BOND _allowed_bonds[NUM_BOND] = {
       { "AG_N",  "N",  "Ag", 1.85, 2.70, 2 },
       { "AG_O",  "O",  "Ag", 1.85, 2.70, 2 },
       { "AG_S",  "S",  "Ag", 2.00, 3.00, 2 },
       { "AL_H",  "Al", "H",  1.35, 1.65, 1 },
       { "AS_H",  "As", "H",  1.20, 1.60, 1 },
       { "AU_N",  "N",  "Au", 1.80, 2.80, 2 },
       { "AU_O",  "O",  "Au", 1.80, 2.80, 2 },
       { "AU_S",  "S",  "Au", 1.80, 3.00, 2 },
       { "BA_O",  "O",  "Ba", 2.50, 3.50, 2 },
       { "B_B",   "B",  "B",  1.45, 1.95, 1 },
       { "B_C",   "C",  "B",  1.20, 1.85, 1 },
       { "BE_N",  "N",  "Be", 1.40, 2.20, 2 },
       { "BE_O",  "O",  "Be", 1.40, 2.20, 2 },
       { "B_F",   "F",  "B",  1.20, 1.75, 1 },
       { "B_N",   "N",  "B",  1.20, 1.65, 1 },
       { "B_O",   "O",  "B",  1.20, 1.65, 1 },
       { "BR_C",  "C",  "Br", 1.70, 2.15, 1 },
       { "BR_N",  "N",  "Br", 1.70, 2.15, 1 },
       { "BR_O",  "O",  "Br", 1.50, 2.00, 1 },
       { "CA_O",  "O",  "Ca", 1.85, 3.20, 2 },
       { "C_C",   "C",  "C",  1.05, 1.65, 1 },
       { "C_CL",  "C",  "Cl", 1.50, 1.95, 1 },
       { "C_CO",  "C",  "Co", 1.70, 2.80, 2 },
       { "CD_N",  "N",  "Cd", 1.75, 2.70, 2 },
       { "CD_O",  "O",  "Cd", 1.75, 2.70, 2 },
       { "CD_S",  "S",  "Cd", 1.90, 3.10, 2 },
       { "CE_N",  "N",  "Ce", 2.00, 3.50, 2 },
       { "CE_O",  "O",  "Ce", 2.00, 3.50, 2 },
       { "C_F",   "C",  "F",  1.20, 1.65, 1 },
       { "C_H",   "C",  "H",  0.85, 1.15, 1 },
       { "C_I",   "C",  "I",  1.95, 2.30, 1 },
       { "CL_CU", "Cl", "Cu", 2.10, 2.80, 2 },
       { "CL_MN", "Mn", "Cl", 1.90, 2.80, 2 },
       { "CL_NI", "Cl", "Ni", 1.90, 2.70, 2 },
       { "CL_N",  "N",  "Cl", 1.50, 1.95, 1 },
       { "CL_O",  "O",  "Cl", 1.40, 1.95, 1 },
       { "CL_RE", "Cl", "Re", 1.90, 2.70, 2 },
       { "C_N",   "C",  "N",  1.20, 1.60, 1 },
       { "C_O",   "C",  "O",  1.10, 1.55, 1 },
       { "CO_N",  "N",  "Co", 1.70, 2.80, 2 },
       { "CO_O",  "O",  "Co", 1.70, 2.80, 2 },
       { "C_P",   "C",  "P",  1.65, 1.87, 1 },
       { "C_RE",  "C",  "Re", 1.80, 2.80, 2 },
       { "CR_N",  "N",  "Cr", 1.70, 2.80, 2 },
       { "CR_O",  "O",  "Cr", 1.70, 2.80, 2 },
       { "C_S",   "C",  "S",  1.55, 2.00, 1 },
       { "C_SE",  "C",  "Se", 1.75, 2.15, 1 },
       { "CS_N",  "N",  "Cs", 2.30, 3.30, 2 },
       { "CS_O",  "O",  "Cs", 2.30, 3.50, 2 },
       { "CU_CU", "Cu", "Cu", 1.80, 2.80, 2 },
       { "CU_N",  "N",  "Cu", 1.70, 2.70, 2 },
       { "CU_O",  "O",  "Cu", 1.70, 2.70, 2 },
       { "CU_S",  "S",  "Cu", 1.90, 2.70, 2 },
       { "DY_N",  "N",  "Dy", 2.00, 3.50, 2 },
       { "DY_O",  "O",  "Dy", 2.00, 3.50, 2 },
       { "ER_N",  "N",  "Er", 2.00, 3.50, 2 },
       { "ER_O",  "O",  "Er", 2.00, 3.50, 2 },
       { "EU_N",  "N",  "Eu", 2.00, 3.50, 2 },
       { "EU_O",  "O",  "Eu", 2.00, 3.50, 2 },
       { "FE_N",  "N",  "Fe", 1.70, 2.80, 2 },
       { "FE_O",  "O",  "Fe", 1.70, 2.80, 2 },
       { "FE_P",  "P",  "Fe", 1.90, 2.70, 2 },
       { "FE_SE", "Se", "Fe", 2.10, 2.70, 2 },
       { "FE_S",  "S",  "Fe", 1.90, 2.70, 2 },
       { "F_MN",  "Mn", "F",  1.60, 2.50, 2 },
       { "F_N",   "N",  "F",  1.10, 1.60, 1 },
       { "F_P",   "P",  "F",  1.25, 1.75, 1 },
       { "GA_N",  "N",  "Ga", 1.70, 2.30, 2 },
       { "GA_O",  "O",  "Ga", 1.70, 2.30, 2 },
       { "GA_S",  "S",  "Ga", 2.00, 2.60, 2 },
       { "GD_N",  "N",  "Gd", 2.00, 3.50, 2 },
       { "GD_O",  "O",  "Gd", 2.00, 3.50, 2 },
       { "HG_N",  "N",  "Hg", 2.00, 2.80, 2 },
       { "HG_O",  "O",  "Hg", 2.00, 3.20, 2 },
       { "HG_S",  "S",  "Hg", 2.30, 3.30, 2 },
       { "H_N",   "N",  "H",  0.85, 1.15, 1 },
       { "HO_N",  "N",  "Ho", 2.00, 3.50, 2 },
       { "H_O",   "O",  "H",  0.85, 1.15, 1 },
       { "HO_O",  "O",  "Ho", 2.00, 3.50, 2 },
       { "H_P",   "P",  "H",  1.10, 1.45, 1 },
       { "H_SI",  "Si", "H",  1.20, 1.60, 1 },
       { "H_S",   "S",  "H",  1.10, 1.45, 1 },
       { "I_I",   "I",  "I",  2.70, 3.10, 1 },
       { "I_N",   "N",  "I",  1.95, 2.30, 1 },
       { "IN_N",  "N",  "In", 2.00, 2.70, 2 },
       { "IN_O",  "O",  "In", 2.00, 2.70, 2 },
       { "I_O",   "O",  "I",  1.95, 2.30, 1 },
       { "IR_N",  "N",  "Ir", 1.70, 2.40, 2 },
       { "IR_O",  "O",  "Ir", 1.70, 2.40, 2 },
       { "I_S",   "S",  "I",  2.70, 3.10, 1 },
       { "K_O",   "O",  "K",  2.00, 3.50, 2 },
       { "LA_N",  "N",  "La", 2.00, 3.50, 2 },
       { "LA_O",  "O",  "La", 2.00, 3.50, 2 },
       { "LI_O",  "O",  "Li", 1.70, 2.60, 2 },
       { "LU_N",  "N",  "Lu", 2.00, 3.50, 2 },
       { "LU_O",  "O",  "Lu", 2.00, 3.50, 2 },
       { "MG_O",  "O",  "Mg", 1.70, 3.00, 2 },
       { "MN_MN", "Mn", "Mn", 2.80, 3.00, 2 },
       { "MN_N",  "Mn", "N",  1.70, 2.80, 2 },
       { "MN_O",  "Mn", "O",  1.70, 2.80, 2 },
       { "MN_P",  "Mn", "P",  1.90, 2.60, 2 },
       { "MN_SE", "Mn", "Se", 2.20, 2.90, 2 },
       { "MN_S",  "Mn", "S",  2.00, 2.90, 2 },
       { "MO_P",  "Mo", "P",  2.10, 2.80, 2 },
       { "MO_SE", "Mo", "Se", 2.10, 2.90, 2 },
       { "MO_S",  "Mo", "S",  2.00, 2.90, 2 },
       { "NA_O",  "O",  "Na", 1.90, 3.20, 2 },
       { "NI_O",  "O",  "Ni", 1.70, 2.80, 2 },
       { "NI_S",  "S",  "Ni", 1.90, 2.70, 2 },
       { "N_NI",  "N",  "Ni", 1.70, 2.80, 2 },
       { "N_N",   "N",  "N",  1.15, 1.55, 1 },
       { "N_O",   "N",  "O",  1.18, 1.50, 1 },
       { "N_OS",  "N",  "Os", 1.70, 2.80, 2 },
       { "N_PB",  "N",  "Pb", 1.90, 3.30, 2 },
       { "N_PD",  "N",  "Pd", 1.80, 2.80, 2 },
       { "N_P",   "N",  "P",  1.54, 1.77, 1 },
       { "N_PR",  "N",  "Pr", 2.00, 3.50, 2 },
       { "N_PT",  "N",  "Pt", 1.80, 2.80, 2 },
       { "N_RB",  "N",  "Rb", 2.40, 3.20, 2 },
       { "N_RE",  "N",  "Re", 1.80, 2.80, 2 },
       { "N_RH",  "N",  "Rh", 1.70, 2.80, 2 },
       { "N_RU",  "N",  "Ru", 1.70, 2.80, 2 },
       { "N_SM",  "N",  "Sm", 2.00, 3.50, 2 },
       { "N_S",   "N",  "S",  1.45, 1.90, 1 },
       { "N_SR",  "N",  "Sr", 2.45, 3.00, 2 },
       { "N_TB",  "N",  "Tb", 2.00, 3.50, 2 },
       { "N_TL",  "N",  "Tl", 2.20, 3.60, 2 },
       { "N_U",   "N",  "U",  1.90, 3.90, 2 },
       { "N_V",   "N",  "V",  1.70, 2.80, 2 },
       { "N_W",   "N",  "W",  2.00, 3.30, 2 },
       { "N_YB",  "N",  "Yb", 2.00, 3.50, 2 },
       { "N_Y",   "N",  "Y",  2.00, 3.50, 2 },
       { "N_ZN",  "N",  "Zn", 1.70, 2.70, 2 },
       { "O_O",   "O",  "O",  1.30, 1.60, 1 },
       { "O_OS",  "O",  "Os", 1.70, 2.80, 2 },
       { "O_PB",  "O",  "Pb", 1.90, 3.30, 2 },
       { "O_PD",  "O",  "Pd", 1.80, 2.80, 2 },
       { "O_P",   "O",  "P",  1.40, 1.75, 1 },
       { "O_PR",  "O",  "Pr", 2.00, 3.50, 2 },
       { "O_PT",  "O",  "Pt", 1.80, 2.80, 2 },
       { "O_RB",  "O",  "Rb", 2.40, 3.20, 2 },
       { "O_RE",  "O",  "Re", 1.80, 2.80, 2 },
       { "O_RH",  "O",  "Rh", 1.70, 2.80, 2 },
       { "O_RU",  "O",  "Ru", 1.70, 2.80, 2 },
       { "O_SM",  "O",  "Sm", 2.00, 3.50, 2 },
       { "O_S",   "O",  "S",  1.35, 1.80, 1 },
       { "O_SR",  "O",  "Sr", 2.30, 3.00, 2 },
       { "O_TB",  "O",  "Tb", 2.00, 3.50, 2 },
       { "O_TE",  "O",  "Te", 1.80, 2.50, 2 },
       { "O_TL",  "O",  "Tl", 2.20, 3.60, 2 },
       { "O_U",   "O",  "U",  1.90, 3.90, 2 },
       { "O_V",   "O",  "V",  1.70, 2.80, 2 },
       { "O_W",   "O",  "W",  2.00, 3.30, 2 },
       { "O_YB",  "O",  "Yb", 2.00, 3.50, 2 },
       { "O_Y",   "O",  "Y",  2.00, 3.50, 2 },
       { "O_ZN",  "O",  "Zn", 1.70, 2.70, 2 },
       { "PD_S",  "S",  "Pd", 2.00, 2.80, 2 },
       { "P_P",   "P",  "P",  1.95, 2.35, 1 },
       { "P_RE",  "P",  "Re", 2.00, 2.80, 2 },
       { "PT_S",  "S",  "Pt", 2.00, 2.90, 2 },
       { "RE_S",  "S",  "Re", 2.10, 2.80, 2 },
       { "RU_S",  "S",  "Ru", 2.20, 3.20, 2 },
       { "SE_SE", "Se", "Se", 2.05, 2.55, 1 },
       { "S_SE",  "S",  "Se", 1.95, 2.40, 1 },
       { "S_S",   "S",  "S",  1.85, 2.20, 1 },
       { "S_W",   "S",  "W",  2.00, 2.80, 2 },
       { "S_ZN",  "S",  "Zn", 1.90, 3.00, 2 }
};

static std::map<std::string, int> _bondIndex;

static void get_bond_index()
{
       if (!_bondIndex.empty()) return;

       std::string cs;
       for (int i = 0; i < NUM_BOND; i++) {
            cs = _allowed_bonds[i].label;
            _bondIndex.insert(make_pair(cs, i));
       }
}

static _ALLOWED_BOND *find_bond(const std::string& label)
{
       get_bond_index();
       std::map<std::string, int>::iterator pos = _bondIndex.find(label);
       if (pos != _bondIndex.end())
            return &_allowed_bonds[pos->second];
       else return NULL;
}
