#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "_maxit.h"
#include "assembly_parser_util.h"
#include "erf_util.h"

#define NUM_ITEM   12

static const char *matrix_items[NUM_ITEM] = {
       "matrix[1][1]", "matrix[1][2]", "matrix[1][3]", "vector[1]",
       "matrix[2][1]", "matrix[2][2]", "matrix[2][3]", "vector[2]",
       "matrix[3][1]", "matrix[3][2]", "matrix[3][3]", "vector[3]"
};

static void init_identical_matrix(double X[4][4]);
static bool is_identical_matrix(const double X[4][4]);
static void mult_vector(const double matrix[4][4], const double vector[4], double result[4]);
static void mult_matrices(const double matrix1[4][4], const double matrix2[4][4], double result[4][4]);
static void get_matrix_from_table(const int row, ISTable *t, double matrix[4][4]);
static void update_matrix_to_table(const int row, ISTable *t, const double matrix[4][4]);
static void MatrixInversion(const double x[4][4], double rx[4][4]);
static void MatrixInversion(double **A, int order, double **Y);
static int GetMinor(double **src, double **dest, int row, int col, int order);
static double CalcDeterminant( double **mat, int order);


void _MAXIT::move_to_crystal_frame()
{
       ISTable *t = _CifObj->GetTablePtr(_StructureId, "pdbx_struct_assembly_gen");
       if (is_empty_table(t)) return;
      
       CifString cs, oper_expression;
       oper_expression.Clear();
       int rowNo = t->GetNumRows();
       int index1 = t->GetColumnIndex("assembly_id");
       int index2 = t->GetColumnIndex("oper_expression");
       for (int i = 0; i < rowNo; ++i) {
            get_value(cs, t, index1, i);
            if (cs == "XAU") {
                 get_value(oper_expression, t, index2, i);
                 break;
            }
       }
       if (oper_expression == "") return;

       char buffer[1000];

       strcpy(buffer, oper_expression.Text());
       char *p = strstr(buffer, "(X0)");
       if (p) oper_expression.Copy(p + 4);

       t = _CifObj->GetTablePtr(_StructureId, "pdbx_struct_oper_list");
       if (is_empty_table(t)) return;

       double X[4][4], rX[4][4];
       init_identical_matrix(X);

       rowNo = t->GetNumRows();
       index1 = t->GetColumnIndex("id");
       for (int i = 0; i < rowNo; i++) {
            get_value(cs, t, index1, i);
            if (cs == "X0") {
                 get_matrix_from_table(i, t, X);
                 break;
            }
       }

       MatrixInversion(X, rX);

       ReVarCifArray<CifString> oplist, id_list;
       oplist.Clear();
       parseString(oper_expression, oplist);
       if (oplist.Length() > 0) {
            ISTable *t1 = new ISTable("struct_ncs_oper");
            t1->AddColumn("id");
            t1->AddColumn("code");
            t1->AddColumn("details ");
            t1->AddColumn("matrix[1][1]");
            t1->AddColumn("matrix[1][2]");
            t1->AddColumn("matrix[1][3]");
            t1->AddColumn("matrix[2][1]");
            t1->AddColumn("matrix[2][2]");
            t1->AddColumn("matrix[2][3]");
            t1->AddColumn("matrix[3][1]");
            t1->AddColumn("matrix[3][2]");
            t1->AddColumn("matrix[3][3]");
            t1->AddColumn("vector[1]");
            t1->AddColumn("vector[2]");
            t1->AddColumn("vector[3]");

            double mat1[4][4], mat2[4][4];
            init_identical_matrix(mat1);
            int row = 0;
            t1->AddRow();
            sprintf(buffer, "%d", row + 1);
            cs.Copy(buffer);
            t1->UpdateCell(cs, 0, row);
            cs.Copy("given");
            t1->UpdateCell(cs, 1, row);
            update_matrix_to_table(row, t1, mat1);
            row++;
            for (unsigned int i = 0; i < oplist.Length(); ++i) {
                 get_multi_matrix_list(oplist[i].Text(), id_list);
                 if (get_final_matrix(mat1, id_list, t)) {
                      if (is_identical_matrix(mat1)) continue;
                      mult_matrices(X, mat1, mat2);
                      mult_matrices(mat2, rX, mat1);
                      t1->AddRow();
                      sprintf(buffer, "%d", row + 1);
                      cs.Copy(buffer);
                      t1->UpdateCell(cs, 0, row);
                      cs.Copy("generate");
                      t1->UpdateCell(cs, 1, row);
                      update_matrix_to_table(row, t1, mat1);
                      row++;
                 }
            }
            _CifObj->WriteTable(t1, _StructureId, "struct_ncs_oper");
       }

       t = _CifObj->GetTablePtr(_StructureId, "atom_site");
       rowNo = t->GetNumRows();
       double vec1[4], vec2[4];
       vec1[3] = 0.0;
       index1 = t->GetColumnIndex("Cartn_x");
       index2 = t->GetColumnIndex("Cartn_y");
       int index3 = t->GetColumnIndex("Cartn_z");
       for (int i = 0; i < rowNo; i++) {
            get_value(cs, t, index1, i); vec1[0] = atof(cs.Text());
            get_value(cs, t, index2, i); vec1[1] = atof(cs.Text());
            get_value(cs, t, index3, i); vec1[2] = atof(cs.Text());
            mult_vector(X, vec1, vec2);
            sprintf(buffer, "%.3f", vec2[0]);
            cs.Copy(buffer);
            t->UpdateCell(cs, index1, i);
            sprintf(buffer, "%.3f", vec2[1]);
            cs.Copy(buffer);
            t->UpdateCell(cs, index2, i);
            sprintf(buffer, "%.3f", vec2[2]);
            cs.Copy(buffer);
            t->UpdateCell(cs, index3, i);
       }
       _CifObj->WriteTable(t, _StructureId, "atom_site");
}

static void init_identical_matrix(double X[4][4])
{
       X[0][0] = X[1][1] = X[2][2] = X[3][3] = 1.0;
       X[0][1] = X[0][2] = X[0][3] = X[1][0] = X[1][2] = X[1][3] =
       X[2][0] = X[2][1] = X[2][3] = X[3][0] = X[3][1] = X[3][2] = 0.0;
}

static bool is_identical_matrix(const double X[4][4])
{
       bool is_idential = true;
       for (int i = 0; i < 4; ++i) {
            for (int j = 0; j < 4; ++j) {
                 if (i == j) {
                      if (fabs(X[i][j] - 1.0) > 0.00001) {
                           is_idential = false;
                           break;
                      }
                 } else if (fabs(X[i][j]) > 0.00001) {
                      is_idential = false;
                      break;
                 }
            }
            if (!is_idential) break;
       }
       return is_idential;
}

static void mult_vector(const double matrix[4][4], const double vector[4], double result[4])
{
       for (int i = 0; i < 4; ++i) {
            result[i] = 0.0;
            for (int j = 0; j < 4; ++j) {
                 result[i] += matrix[i][j] * vector[j];
            }
       }
}

static void mult_matrices(const double matrix1[4][4], const double matrix2[4][4], double result[4][4])
{
       for (int i = 0; i < 4; ++i) {
            for (int j = 0; j < 4; ++j) {
                 result[i][j] = 0.0;
                 for (int k = 0; k < 4; ++k)
                      result[i][j] += (matrix1[i][k] * matrix2[k][j]);
            }
       }
}

static void get_matrix_from_table(const int row, ISTable *t, double matrix[4][4])
{
       CifString cs;
       for (int i = 0; i < 3; ++i) {
            for (int j = 0; j < 4; ++j) {
                 int k = i * 4 + j;
                 int index = t->GetColumnIndex(matrix_items[k]);
                 get_value(cs, t, index, row);
                 matrix[i][j] = atof(cs.Text());
            }
       }
       matrix[3][0] = matrix[3][1] = matrix[3][2] = 0;
       matrix[3][3] = 1;
}

static void update_matrix_to_table(const int row, ISTable *t, const double matrix[4][4])
{
       CifString cs;
       char buffer[20];
       for (int i = 0; i < 3; ++i) {
            for (int j = 0; j < 4; ++j) {
                 int k = i * 4 + j;
                 int index = t->GetColumnIndex(matrix_items[k]);
                 sprintf(buffer, "%.10f", matrix[i][j]);
                 cs.Copy(buffer);
                 t->UpdateCell(cs, index, row);
            }
       }
}

static void MatrixInversion(const double x[4][4], double rx[4][4])
{
       double **a = new double*[4];
       double **b = new double*[4];
       for (int i = 0; i < 4; ++i) {
            a[i] = new double[4];
            b[i] = new double[4];
       }
       for (int i = 0; i < 4; ++i) {
            for (int j = 0; j < 4; ++j) {
                 a[i][j] = x[i][j];
            }
       }
       MatrixInversion(a, 4, b);

       for (int i = 0; i < 4; ++i) {
            for (int j = 0; j < 4; ++j) {
                 rx[i][j] = b[i][j];
            }
       }

       for (int i = 0; i < 4; ++i) {
            delete [] a[i];
            delete [] b[i];
       }
       delete [] a;
       delete [] b;
}

// the result is put in Y
static void MatrixInversion(double **A, int order, double **Y)
{
       // get the determinant of a
       double det = 1.0/CalcDeterminant(A,order);

       // memory allocation
       double *temp = new double[(order-1)*(order-1)];
       double **minor = new double*[order-1];
       for (int i=0;i<order-1;i++) minor[i] = temp+(i*(order-1));

       for (int j=0;j<order;j++) {
            for (int i=0;i<order;i++) {
                 // get the co-factor (matrix) of A(j,i)
                 GetMinor(A,minor,j,i,order);
                 Y[i][j] = det*CalcDeterminant(minor,order-1);
                 if( (i+j)%2 == 1) Y[i][j] = -Y[i][j];
            }
       }

       // release memory
       delete [] minor[0];
       delete [] minor;
}

// calculate the cofactor of element (row,col)
static int GetMinor(double **src, double **dest, int row, int col, int order)
{
       // indicate which col and row is being copied to dest
       int colCount=0,rowCount=0;

       for (int i = 0; i < order; i++ ) {
            if ( i != row ) {
                 colCount = 0;
                 for (int j = 0; j < order; j++ ) {
                      // when j is not the element
                      if ( j != col ) {
                           dest[rowCount][colCount] = src[i][j];
                           colCount++;
                      }
                 }
                 rowCount++;
            }
       }

       return 1;
}

// Calculate the determinant recursively.
static double CalcDeterminant( double **mat, int order)
{
       // order must be >= 0
       // stop the recursion when matrix is a single element
       if( order == 1 ) return mat[0][0];

       // the determinant value
       double det = 0;

       // allocate the cofactor matrix
       double **minor;
       minor = new double*[order-1];
       for(int i=0;i<order-1;i++) minor[i] = new double[order-1];

       for (int i = 0; i < order; i++ ) {
            // get minor of element (0,i)
            GetMinor( mat, minor, 0, i , order);
            // the recusion is here!
            det += pow( -1.0, i ) * mat[0][i] * CalcDeterminant( minor,order-1 );
       }

       // release memory
       for(int i=0;i<order-1;i++)
       delete [] minor[i];
       delete [] minor;

       return det;
}
static void MatrixInversion(const double x[4][4], double rx[4][4]);
static void MatrixInversion(double **A, int order, double **Y);
static int GetMinor(double **src, double **dest, int row, int col, int order);
static double CalcDeterminant( double **mat, int order);
