#include <string.h>

#include "assembly_parser_util.h"
#include "erf_util.h"

static void clean_string(CifString &stringbuff);
static void separate_string_by_parenthesis(CifString &cifstring, ReVarCifArray<CifString> &list);
static void separate_string_by_comma(CifString &cifstring, ReVarCifArray<CifString> &list);
static char *get_word(char *stringbuff, CifString &cs);
static void mult_matrixes(const double matrix1[4][4], const double matrix2[4][4], double result[4][4]);
static int get_matrix_from_oper_list(const CifString id_value, ISTable *Table, double matrix[4][4]);

void get_assembly_list(ReVarCifArray<ReVarCifArray<__GENASSEMBLY> > &assembly_lists,
              CifFileObj *fobj, const char *blockId)
{
       assembly_lists.Clear();

       if (!fobj) return;

       if (is_empty_table(fobj->GetTablePtr(blockId, "pdbx_struct_assembly")) ||
           is_empty_table(fobj->GetTablePtr(blockId, "pdbx_struct_assembly_gen")) ||
           is_empty_table(fobj->GetTablePtr(blockId, "pdbx_struct_oper_list"))) {
            return;
       }

       ISTable *t = fobj->GetTablePtr(blockId, "pdbx_struct_assembly");

       CifString cs1, cs2;
       ReVarCifArray<CifString> assembly_id_lists;
       assembly_id_lists.Clear();
       int rowNo = t->GetNumRows();
       int idx1 = t->GetColumnIndex("id");
       int idx2 = t->GetColumnIndex("details");
       for (int i = 0; i < rowNo; i++) {
            get_value(cs2, t, idx2, i);
            cs2.ToLower();
            if (cs2 == "representative helical assembly" ||
                cs2 == "complete icosahedral assembly" ||
                cs2 == "complete point assembly" ||
                cs2 == "author_and_software_defined_assembly" ||
                cs2 == "author_defined_assembly" ||
                cs2 == "software_defined_assembly") {
                 get_value(cs1, t, idx1, i);
                 assembly_id_lists.Add(cs1);
            }
       }
       if (assembly_id_lists.Length() == 0) return;

       ISTable *t1 = fobj->GetTablePtr(blockId, "pdbx_struct_oper_list");
       t = fobj->GetTablePtr(blockId, "pdbx_struct_assembly_gen");
       rowNo = t->GetNumRows();
       idx1 = t->GetColumnIndex("assembly_id");
       idx2 = t->GetColumnIndex("asym_id_list");
       int idx3 = t->GetColumnIndex("oper_expression");
       CifString cs3;
       char *buff = NULL, *p = NULL, buffer[80];
       ReVarCifArray<__GENASSEMBLY> assembly_list;
       ReVarCifArray<CifString> AsymIds, oplist, id_list;
       double mat[4][4];
       for (unsigned int i = 0; i < assembly_id_lists.Length(); i++) {
            assembly_list.Clear();
            for (int j = 0; j < rowNo; j++) {
                 get_value(cs1, t, idx1, j);
                 get_value(cs2, t, idx2, j);
                 get_value(cs3, t, idx3, j);
                 if (cs1 == assembly_id_lists[i]) {
                      int len = strlen(cs2.Text());
                      buff = new char[len + 1];
                      memset(buff, 0, len + 1);
                      strcpy(buff, cs2.Text());
                      for (int k = 0; k < len; k++) {
                           if (buff[k] == ',' || buff[k] == ';') buff[k] = ' ';
                      }
                      ndb_clean_string(buff);

                      AsymIds.Clear();
                      p = buff;
                      while (p) {
                           p = get_word(p, buffer);
                           if (!strcmp(buffer, "")) continue;
                           cs2.Copy(buffer);
                           AsymIds.Add(cs2);
                      }
                      delete [] buff;
    
                      oplist.Clear();
                      parseString(cs3, oplist);
                      if (oplist.Length() == 0 || AsymIds.Length() == 0) continue;

                      for (unsigned int k = 0; k < oplist.Length(); k++) {
                           get_multi_matrix_list(oplist[k].Text(), id_list);
                           if (get_final_matrix(mat, id_list, t1)) {
                                __GENASSEMBLY assembly;
                                assembly.AsymIds = AsymIds;
                                for (int l = 0; l < 3; l++) {
                                     for (int m = 0; m < 4; m++) {
                                          assembly.Matrix[l][m] = mat[l][m];
                                     }
                                }
                                assembly_list.Add(assembly);
                           }
                      }
                 }
            }
            if (assembly_list.Length() > 0) assembly_lists.Add(assembly_list);
       }
}

int parseString(CifString &op_string, ReVarCifArray<CifString> &list)
{
       CifString cs;
       clean_string(op_string);
       int has_special_character = 0;
       for (unsigned int i = 0; i < op_string.Length(); i++) {
            if (op_string[i] == '(' || op_string[i] == ')' ||
                op_string[i] == ',' || op_string[i] == '-') {
                 has_special_character = 1;
                 break;
            }
       }
       if (!has_special_character) {
            list.Add(op_string);
            return 0;
       }

       int numlp = 0;
       int numrp = 0;
       int numd = 0;
       int firstl = -1;
       int firstr = -1;
       int firstc = -1;
      
       for (unsigned int i = 0; i < op_string.Length(); i++) {
            if (op_string[i] == '(') {
                 numlp++;
                 if (firstl < 0) firstl = i;
            } else if (op_string[i] == ')') {
                 numrp++;
                 if (firstr < 0) firstr = i;
            } else if (op_string[i] == ',') {
                 if (firstc < 0) firstc = i;
            } else if (op_string[i] == '-') {
                 numd++;
            }
       } 

       if (numlp != numrp) return 2;
       if (numlp && numrp && firstr < firstl) return 2;

       int choice = 0;
       if (firstl >= 0) {
            if (firstc >= 0 && firstc < firstl)
                 choice = 2;
            else choice = 1;
       } else if (firstc >= 0) {
            if (firstl >= 0 && firstl < firstc)
                 choice = 1;
            else choice = 2;
       } else if (numd == 1) choice = 3;

       ReVarCifArray<CifString> list1, list2;
       if (choice == 1) {
            ReVarCifArray<ReVarCifArray<CifString> > lists;
            lists.Clear();
            separate_string_by_parenthesis(op_string, list1); 
            if (!list1.Length()) return 2;
            for (unsigned int i = 0; i < list1.Length(); i++) {
                 list2.Clear();
                 int error = parseString(list1[i], list2);
                 if (error) return error;
                 lists.Add(list2);
            }
            if (!lists.Length()) return 2;

            list2 = lists[lists.Length() - 1];
            for (int i = (int) lists.Length() - 2; i >= 0; i--) {
                 list1.Clear();
                 for (unsigned int j = 0; j < lists[i].Length(); j++) {
                      for (unsigned int k = 0; k < list2.Length(); k++) {
                           cs.Copy(lists[i][j]);
                           cs += " ";
                           cs += list2[k];
                           list1.Add(cs);
                      }
                 }
                 list2 = list1;
            }
            for (unsigned int i = 0; i < list2.Length(); i++) {
                 list.Add(list2[i]);
            }
            return 0;
       } else if (choice == 2) {
            separate_string_by_comma(op_string, list1);
            if (!list1.Length()) return 3;
            for (unsigned int i = 0; i < list1.Length(); i++) {
                 list2.Clear();
                 int error = parseString(list1[i], list2);
                 if (error) return error;
                 for (unsigned int j = 0; j < list2.Length(); j++) {
                      list.Add(list2[j]);
                 }
            }
            return 0;
       } else if (choice == 3) {
            char *r = strstr(op_string.Text(), "-");
            int start = atoi(op_string.Text());
            int end = atoi(r + 1);
            if (end <= start) return 3;
            if (end == 0) return 3;
            char buf[10];
            for (int i = start; i <= end; i++) {
                 sprintf(buf, "%d", i);
                 cs.Copy(buf);
                 list.Add(cs);
            }
            return 0;
       } else return 4;
}

void get_multi_matrix_list(char *id_string, ReVarCifArray<CifString> &id_list)
{
       id_list.Clear();
       CifString cs;
       char *p = id_string;
       while (p) {
            p = get_word(p, cs);
            if (cs != "") id_list.Add(cs);
       }
}

int get_final_matrix(double mat[4][4], ReVarCifArray<CifString> &id_list, ISTable *Table)
{
       double mat1[4][4], mat2[4][4];

       if (id_list.Length() > 0) {
            if (!get_matrix_from_oper_list(id_list[id_list.Length()-1], Table, mat)) return 0;
            for (int i = (int) id_list.Length() - 2; i >= 0; i--) {
                 if (!get_matrix_from_oper_list(id_list[i], Table, mat1)) return 0;
                 mult_matrixes(mat1, mat, mat2);
                 for (int j = 0; j < 4; j++) {
                      for (int k = 0; k < 4; k++) {
                           mat[j][k] = mat2[j][k];
                      }
                 }
            }
            return 1;
       }
       return 0;
}

static void clean_string(CifString &stringbuff)
{
       CifString cs;

       cs.Clear();
       for (unsigned int i = 0; i < stringbuff.Length(); i++) {
            if (stringbuff[i] != ' ' &&
                stringbuff[i] != '\t' &&
                stringbuff[i] != '\n') {
                 cs += stringbuff[i];
            }
       }
       stringbuff.Copy(cs);
}

static void separate_string_by_parenthesis(CifString &cifstring, ReVarCifArray<CifString> &list)
{
       list.Clear();
       CifString cs;
       cs.Clear();
       int count = 0;
       for (unsigned int i = 0; i < cifstring.Length(); i++) {
            if (cifstring[i] == ' ' || cifstring[i] == '\t' ||
                cifstring[i] == '\n') continue;
            if (cifstring[i] == '(') {
                 if (count != 0) cs += cifstring[i];
                 count++;
            } else if (cifstring[i] == ')') {
                 count--;
                 if (count == 0) {
                      if (cs != "") list.Add(cs);
                      cs.Clear();
                 } else cs += cifstring[i];
            } else cs += cifstring[i];
       }
}

static void separate_string_by_comma(CifString &cifstring, ReVarCifArray<CifString> &list)
{
       list.Clear();
       CifString cs;
       cs.Clear();
       int count = 0;
       for (unsigned int i = 0; i < cifstring.Length(); i++) {
            if (cifstring[i] == ' ' || cifstring[i] == '\t' ||
                cifstring[i] == '\n') continue;
            if (cifstring[i] == ',' && count == 0) {
                 if (cs != "") list.Add(cs);
                 cs.Clear();
                 i++;
            }
            cs += cifstring[i];
            if (cifstring[i] == '(') count++;
            if (cifstring[i] == ')') count--;
       }
       if (cs != "") list.Add(cs);
}

static char *get_word(char *stringbuff, CifString &cs)
{
       cs.Clear();
       while ((*stringbuff == ' ' || *stringbuff == '\t') &&
               *stringbuff != '\0') stringbuff++;
       if (*stringbuff == '\0') return NULL;
       while (*stringbuff != ' ' && *stringbuff != '\t' && *stringbuff != '\0') {
            cs += *stringbuff;
            stringbuff++;
       }
       if (*stringbuff == '\0') return NULL;
       return stringbuff;
}

static void mult_matrixes(const double matrix1[4][4], const double matrix2[4][4], double result[4][4])
{
       int i, j, k;
       for (i = 0; i < 4; i++) {
            for (j = 0; j < 4; j++) {
                 result[i][j] = 0.0;
                 for (k = 0; k < 4; k++)
                      result[i][j] += (matrix1[i][k] * matrix2[k][j]);
            }
       }
}

static int get_matrix_from_oper_list(const CifString id_value, ISTable *Table, double matrix[4][4])
{
       int rowNo = Table->GetNumRows();
       int id  = Table->GetColumnIndex("id");
       int m11 = Table->GetColumnIndex("matrix[1][1]");
       int m12 = Table->GetColumnIndex("matrix[1][2]");
       int m13 = Table->GetColumnIndex("matrix[1][3]");
       int m21 = Table->GetColumnIndex("matrix[2][1]");
       int m22 = Table->GetColumnIndex("matrix[2][2]");
       int m23 = Table->GetColumnIndex("matrix[2][3]");
       int m31 = Table->GetColumnIndex("matrix[3][1]");
       int m32 = Table->GetColumnIndex("matrix[3][2]");
       int m33 = Table->GetColumnIndex("matrix[3][3]");
       int v1  = Table->GetColumnIndex("vector[1]");
       int v2  = Table->GetColumnIndex("vector[2]");
       int v3  = Table->GetColumnIndex("vector[3]");
       CifString cs;
       for (int i = 0; i < rowNo; i++) {
            Table->GetCell(cs, id, i);
            if (cs == id_value) {
                 Table->GetCell(cs, m11, i); matrix[0][0] = atof(cs.Text());
                 Table->GetCell(cs, m12, i); matrix[0][1] = atof(cs.Text());
                 Table->GetCell(cs, m13, i); matrix[0][2] = atof(cs.Text());
                 Table->GetCell(cs, v1, i);  matrix[0][3] = atof(cs.Text());

                 Table->GetCell(cs, m21, i); matrix[1][0] = atof(cs.Text());
                 Table->GetCell(cs, m22, i); matrix[1][1] = atof(cs.Text());
                 Table->GetCell(cs, m23, i); matrix[1][2] = atof(cs.Text());
                 Table->GetCell(cs, v2, i);  matrix[1][3] = atof(cs.Text());

                 Table->GetCell(cs, m31, i); matrix[2][0] = atof(cs.Text());
                 Table->GetCell(cs, m32, i); matrix[2][1] = atof(cs.Text());
                 Table->GetCell(cs, m33, i); matrix[2][2] = atof(cs.Text());
                 Table->GetCell(cs, v3, i);  matrix[2][3] = atof(cs.Text());

                 matrix[3][0] = matrix[3][1] = matrix[3][2] = 0;
                 matrix[3][3] = 1;

                 return 1;
            }
       }

       return 0;
}
