/*
FILE:     write_letter.C
*/
/*
VERSION:  8.200
*/
/*
DATE:     3/18/2015
*/
/*
  Comments and Questions to: sw-help@rcsb.rutgers.edu
*/
/*
COPYRIGHT 1999-2015 Rutgers - The State University of New Jersey

This software is provided WITHOUT WARRANTY OF MERCHANTABILITY OR
FITNESS FOR A PARTICULAR PURPOSE OR ANY OTHER WARRANTY, EXPRESS OR
IMPLIED.  RUTGERS MAKE NO REPRESENTATION OR WARRANTY THAT THE
SOFTWARE WILL NOT INFRINGE ANY PATENT, COPYRIGHT OR OTHER
PROPRIETARY RIGHT.

The user of this software shall indemnify, hold harmless and defend
Rutgers, its governors, trustees, officers, employees, students,
agents and the authors against any and all claims, suits,
losses, liabilities, damages, costs, fees, and expenses including
reasonable attorneys' fees resulting from or arising out of the
use of this software.  This indemnification shall include, but is
not limited to, any and all claims alleging products liability.
*/
/*
               PDB SOFTWARE LICENSE AGREEMENT

BY CLICKING THE ACCEPTANCE BUTTON OR INSTALLING OR USING 
THIS "SOFTWARE, THE INDIVIDUAL OR ENTITY LICENSING THE  
SOFTWARE ("LICENSEE") IS CONSENTING TO BE BOUND BY AND IS 
BECOMING A PARTY TO THIS AGREEMENT.  IF LICENSEE DOES NOT 
AGREE TO ALL OF THE TERMS OF THIS AGREEMENT
THE LICENSEE MUST NOT INSTALL OR USE THE SOFTWARE.

1. LICENSE AGREEMENT

This is a license between you ("Licensee") and the Protein Data Bank (PDB) 
at Rutgers, The State University of New Jersey (hereafter referred to 
as "RUTGERS").   The software is owned by RUTGERS and protected by 
copyright laws, and some elements are protected by laws governing 
trademarks, trade dress and trade secrets, and may be protected by 
patent laws. 

2. LICENSE GRANT

RUTGERS grants you, and you hereby accept, non-exclusive, royalty-free 
perpetual license to install, use, modify, prepare derivative works, 
incorporate into other computer software, and distribute in binary 
and source code format, or any derivative work thereof, together with 
any associated media, printed materials, and on-line or electronic 
documentation (if any) provided by RUTGERS (collectively, the "SOFTWARE"), 
subject to the following terms and conditions: (i) any distribution 
of the SOFTWARE shall bind the receiver to the terms and conditions 
of this Agreement; (ii) any distribution of the SOFTWARE in modified 
form shall clearly state that the SOFTWARE has been modified from 
the version originally obtained from RUTGERS.  

2. COPYRIGHT; RETENTION OF RIGHTS.  

The above license grant is conditioned on the following: (i) you must 
reproduce all copyright notices and other proprietary notices on any 
copies of the SOFTWARE and you must not remove such notices; (ii) in 
the event you compile the SOFTWARE, you will include the copyright 
notice with the binary in such a manner as to allow it to be easily 
viewable; (iii) if you incorporate the SOFTWARE into other code, you 
must provide notice that the code contains the SOFTWARE and include 
a copy of the copyright notices and other proprietary notices.  All 
copies of the SOFTWARE shall be subject to the terms of this Agreement.  

3. NO MAINTENANCE OR SUPPORT; TREATMENT OF ENHANCEMENTS 

RUTGERS is under no obligation whatsoever to: (i) provide maintenance 
or support for the SOFTWARE; or (ii) to notify you of bug fixes, patches, 
or upgrades to the features, functionality or performance of the 
SOFTWARE ("Enhancements") (if any), whether developed by RUTGERS 
or third parties.  If, in its sole discretion, RUTGERS makes an 
Enhancement available to you and RUTGERS does not separately enter 
into a written license agreement with you relating to such bug fix, 
patch or upgrade, then it shall be deemed incorporated into the SOFTWARE 
and subject to this Agreement. You are under no obligation whatsoever 
to provide any Enhancements to RUTGERS or the public that you may 
develop over time; however, if you choose to provide your Enhancements 
to RUTGERS, or if you choose to otherwise publish or distribute your 
Enhancements, in source code form without contemporaneously requiring 
end users or RUTGERS to enter into a separate written license agreement 
for such Enhancements, then you hereby grant RUTGERS a non-exclusive,
royalty-free perpetual license to install, use, modify, prepare
derivative works, incorporate into the SOFTWARE or other computer
software, distribute, and sublicense your Enhancements or derivative
works thereof, in binary and source code form.

4. FEES.  There is no license fee for the SOFTWARE.  If Licensee
wishes to receive the SOFTWARE on media, there may be a small charge
for the media and for shipping and handling.  Licensee is
responsible for any and all taxes.

5. TERMINATION.  Without prejudice to any other rights, Licensor
may terminate this Agreement if Licensee breaches any of its terms
and conditions.  Upon termination, Licensee shall destroy all
copies of the SOFTWARE.

6. PROPRIETARY RIGHTS.  Title, ownership rights, and intellectual
property rights in the Product shall remain with RUTGERS.  Licensee 
acknowledges such ownership and intellectual property rights and will 
not take any action to jeopardize, limit or interfere in any manner 
with RUTGERS' ownership of or rights with respect to the SOFTWARE.  
The SOFTWARE is protected by copyright and other intellectual 
property laws and by international treaties.  Title and related 
rights in the content accessed through the SOFTWARE is the property 
of the applicable content owner and is protected by applicable law.  
The license granted under this Agreement gives Licensee no rights to such
content.

7. DISCLAIMER OF WARRANTY.  THE SOFTWARE IS PROVIDED FREE OF 
CHARGE, AND, THEREFORE, ON AN "AS IS" BASIS, WITHOUT WARRANTY OF 
ANY KIND, INCLUDING WITHOUT LIMITATION THE WARRANTIES THAT IT 
IS FREE OF DEFECTS, MERCHANTABLE, FIT FOR A PARTICULAR PURPOSE 
OR NON-INFRINGING.  THE ENTIRE RISK AS TO THE QUALITY AND 
PERFORMANCE OF THE SOFTWARE IS BORNE BY LICENSEE.  SHOULD THE 
SOFTWARE PROVE DEFECTIVE IN ANY RESPECT, THE LICENSEE AND NOT 
LICENSOR ASSUMES THE ENTIRE COST OF ANY SERVICE AND REPAIR.  
THIS DISCLAIMER OF WARRANTY CONSTITUTES AN ESSENTIAL PART OF 
THIS AGREEMENT.  NO USE OF THE PRODUCT IS AUTHORIZED HEREUNDER 
EXCEPT UNDER THIS DISCLAIMER.

8. LIMITATION OF LIABILITY.  TO THE MAXIMUM EXTENT PERMITTED BY
APPLICABLE LAW,  IN NO EVENT WILL LICENSOR BE LIABLE FOR ANY 
INDIRECT, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES ARISING 
OUT OF THE USE OF OR INABILITY TO USE THE SOFTWARE, INCLUDING, 
WITHOUT LIMITATION, DAMAGES FOR LOSS OF GOODWILL, WORK 
STOPPAGE, COMPUTER FAILURE OR MALFUNCTION, OR ANY AND ALL 
OTHER COMMERCIAL DAMAGES OR LOSSES, EVEN IF ADVISED OF THE
POSSIBILITY THEREOF. 
*/
#include <sys/types.h>
#include <time.h>
#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>
#include <math.h>

#include <set>
#include <string>
#include <map>
#include <vector>

#include "_fileobj_ioutil.h"
#include "_validation.h"
#include "erf_util.h"
#include "ndb_misclib.h"
#include "pdbx_remark_500.h"
#include "util.h"

#define NDB_FIND_MISSING_RESIDUE               0
#define NDB_FIND_MISSING_ATOM                  1
#define NDB_FIND_EXTRA_ATOM                    2

#define letterpdb   "pdb_letter1"
#define letternmr   "nmr_letter"

#define letterVSRVxray   "vsrv_xray_letter"
#define letterVSRVnmr    "vsrv_nmr_letter"

#define letterADITxray  "adit_xray_letter"
#define letterADITnmr   "adit_nmr_letter"

typedef struct {
      const char *res;
      const char *t_i;
      const char *r_s;
} _CBETA_CHIRAL;

#define NUM_CBETA_CHIRAL   21

static const _CBETA_CHIRAL _cbeta_chiral[NUM_CBETA_CHIRAL] = {
       { "4TP", "T", "R" },
       { "ALO", "T", "S" },
       { "B2I", "I", "S" },
       { "BIU", "I", "S" },
       { "BMT", "T", "R" },
       { "CTH", "T", "R" },
       { "D11", "T", "S" },
       { "DIL", "I", "R" },
       { "DTH", "T", "S" },
       { "IIL", "I", "R" },
       { "ILE", "I", "S" },
       { "ILP", "I", "S" },
       { "ILX", "I", "S" },
       { "IML", "I", "S" },
       { "LNT", "T", "R" },
       { "OLT", "T", "R" },
       { "TH5", "T", "R" },
       { "THC", "T", "R" },
       { "THO", "T", "R" },
       { "THR", "T", "R" },
       { "TPO", "T", "R" }
};

static std::map<std::string, int> _cbeta_index;

#define CLOSE_CUTOFF  1.0
#define BOND_CUTOFF   0.4
#define ANGLE_CUTOFF  25.0
#define CHIRAL_CUTOFF  "WRONG HAND"
#define CHIRAL_CUTOFF1 "Expecting L Found D"
#define CHIRAL_CUTOFF2 "Expecting D Found L"

static CifString _closeContact;
static CifString _closeHeader;
static CifString _chirality;
static CifString _bond_angle_torsion;
static CifString _backbone;
static CifString _conflict;
static bool _generate_flag = false;

static int  find_letter_token(char *line, char* token);
void ndb_writing_missing_words(FILE *infile, FILE *outfile);
static void check_calpha_chiral_error(CifString &cs, _TORSION *torsion);
static void check_cbeta_chiral_error(CifString &cs, _TORSION *torsion);
static void check_c1na_chiral_error(CifString &cs, _TORSION *torsion);
static void passing_missing_words(const int missing, FILE *infile, const char *oldtoken);
static int is_terminal_atom(_ATOM *atom, _RESIDUE *residue, const char *atom_name);
static float cal_dist(_ATOM *atom_a, _ATOM *atom_b, const double& cutoff);
static bool get_connections(const std::vector<_RESIDUE*>& res_list,
                            std::vector<std::vector<unsigned int> >& connects);
static void write_sugar_linkages(FILE *fp, const std::vector<std::vector<unsigned int> >&
                                 linkages, const std::vector<_RESIDUE*>& res_list);
static void get_linkages(const std::vector<std::vector<unsigned int> >& connects,
                         std::vector<std::vector<std::vector<unsigned int> > >& linkages);
static void clustering(const std::vector<std::vector<unsigned int> >& connects,
                       std::vector<std::set<unsigned int> >& groups);
static void find_linkages(std::vector<std::vector<std::pair<unsigned int, bool> > >&
                          connects, std::vector<std::vector<unsigned int> >& linkages);

void _VALIDATION::get_letter(FILE *outfile, FILE *warning, const char *File_ID, const int is_prev_check,
                             const int public_flag, const int include_hydrogen)
{
       unsigned int i, j;
       int ret;
       char line[MxLen], token[MxLen];
       char date_string[MxStrLen], letterdbname[200];
       time_t curr_time;
       CifString message;
     
       time(&curr_time);
       memset(date_string, 0, MxStrLen);
       strcpy(date_string, ctime(&curr_time));
       strcpy(date_string, &date_string[4]);

       for (i = 0; i < strlen(date_string); i++) {
            if (!strncmp(&date_string[i], "199", 3)) {
                 strcpy(&date_string[7], &date_string[i]);
            }
       } 

       ndb_clean_string(date_string);
       for (i = 0; i < strlen(date_string); i++) {
            if (date_string[i] == ' ') break;
       }
       for (j = strlen(date_string) - 1; j >= i; j--) {
            date_string[j + 1] = date_string[j];
       }
       date_string[i] = '.';
     
       for (i += 2; i < strlen(date_string); i++) {
            if (date_string[i] == ' ') break;
       }
       for (j = strlen(date_string); j >= i; j--) {
            date_string[j + 1] = date_string[j];
       }
       date_string[i] = ',';
/*
       if (_experiment_type & EXPERIMENT_TYPE_NMR || include_hydrogen ||
           _experiment_type & EXPERIMENT_TYPE_NMR_SOLID)
            calculate_all_contact(3.5, 0);
       else calculate_all_contact(3.5, 1);
*/
       calculate_all_contact(WATER_WATER_LIMIT, 0, 3);
       // calculate_all_contact(WATER_LIMIT, 0, 3); // for Kim's method

       finding_missing_and_extra_atoms(0);
/*
       if (public_flag)
            write_validation_letter_for_public_server(outfile);
       else {
*/
/*
            if (_experiment_type & EXPERIMENT_TYPE_NMR ||
                _experiment_type & EXPERIMENT_TYPE_NMR_SOLID ||
                _experiment_type & EXPERIMENT_TYPE_MODEL) {
*/
            if (_cell->is_artifical) {
                 if (is_prev_check == 2)
                      sprintf(letterdbname, "%s/data/ascii/%s", root_dir, letterADITnmr);
                 else if (is_prev_check == 1)
                      sprintf(letterdbname, "%s/data/ascii/%s", root_dir, letterVSRVnmr);
                 else sprintf(letterdbname, "%s/data/ascii/%s", root_dir, letternmr);
            } else {
                 if (is_prev_check==2)
                      sprintf(letterdbname, "%s/data/ascii/%s", root_dir, letterADITxray);
                 else if (is_prev_check == 1)
                      sprintf(letterdbname, "%s/data/ascii/%s", root_dir, letterVSRVxray);
                 else sprintf(letterdbname, "%s/data/ascii/%s", root_dir, letterpdb);
            }

            _closeContact.Clear();
            _closeHeader.Clear();
            _chirality.Clear();
            _bond_angle_torsion.Clear();
            _backbone.Clear();
            _conflict.Clear();
            if (is_prev_check == 2) {
                 _generate_flag = true;
                 char filename[100];
                 sprintf(filename, "%s.keywds", _StructureId);
                 ndb_set_string_to_lower(filename);
                 FILE *fp = fopen(filename, "w");
                 int card_id = find_card_id("KEYWDS");
                 One_Card_Info_Format *pCard = _CardInfo[card_id].Head();
                 if (pCard && strcmp(pCard->Field(2), ""))
                      fprintf(fp, "%s\n", pCard->Field(2));
                 else if (_CifObj) {
                      ISTable *t = _CifObj->GetTablePtr(_StructureId, "struct_keywords");
                      if (!is_empty_table(t)) {
                           CifString cs;
                           int index = t->GetColumnIndex("text");
                           get_value(cs, t, index, 0);
                           fprintf(fp, "%s\n", cs.Text());
                      }
                 }
                 fclose (fp);
            }

            FILE *fp = fopen(letterdbname, "r");
            if (fp) {
                 while (!feof(fp)) {
                      if (fgets(line,MxLen,fp));
                      for (i = 0; i < strlen(line); i++)  {
                           ret = find_letter_token(&line[i], token);
          
                           if (ret == FALSE) fprintf(outfile, "%c", line[i]);
                           else {
                                i += ret;
                                if (!strcmp(token, "DATE")) 
                                     fprintf(outfile, "%s", date_string);
                                else if (!strcmp(token, "REPRINTS"))
                                     passing_missing_words(FALSE, fp, token);
                                else {
                                     ret = write_letter_token(fp, outfile, token);
                                     if (ret == TRUE) break;
                                }
                           }
                      }
                 }
                 fclose(fp);
            } else {
                 ndb_log_message(NDB_MSG_ERR,
                      "Unable to open residue name database: %s\n", letterdbname);
            }
       // }

       check_link_records(warning);
       check_site_records(warning);
       check_terminal_atoms(message);
       if (strcmp(message.Text(), "")) fprintf(warning, "%s\n", message.Text());
       char filename[100];
       if (_closeContact != "") {
            sprintf(filename, "%s.close", File_ID);
            ndb_set_string_to_lower(filename);
            fp = fopen(filename, "w");
            fprintf(fp, "%s\n", _closeContact.Text());
            fclose (fp);
       }
       if (_n_t_pair_outliers && _n_t_pair_checked) {
            double percent = ((double) _n_t_pair_outliers * 100) / ((double) _n_t_pair_checked);
            if (percent >= 6) {
                 sprintf(filename, "%s.torsion", File_ID);
                 ndb_set_string_to_lower(filename);
                 fp = fopen(filename, "w");
                 fprintf(fp, "Total %d pairs were evaluated. %d ( %.2f%% ) were in disallowed regions. The expected\n",
                        _n_t_pair_checked , _n_t_pair_outliers, percent);
                 fprintf(fp, "Ramachandran regions are defined in the following reference: GJ. Kleywegt and TA. Jones,\n");
                 fprintf(fp, "PHI/PSI-chology: Ramachandran Revisited, Structure  1996, 4, 1395 - 1400.\n");
                 fclose (fp);
            }
       }
       if (_chirality != "") {
            sprintf(filename, "%s.chiral", File_ID);
            ndb_set_string_to_lower(filename);
            fp = fopen(filename, "w");
            fprintf(fp, "%s\n", _chirality.Text());
            fclose (fp);
       }
       if (_bond_angle_torsion != "") {
            sprintf(filename, "%s.geom", File_ID);
            ndb_set_string_to_lower(filename);
            fp = fopen(filename, "w");
            fprintf(fp, "%s\n", _bond_angle_torsion.Text());
            fclose (fp);
       }
       if (_backbone != "") {
            sprintf(filename, "%s.backbone", File_ID);
            ndb_set_string_to_lower(filename);
            fp = fopen(filename, "w");
            fprintf(fp, "%s\n", _backbone.Text());
            fclose (fp);
       }
       if (_conflict != "") {
            sprintf(filename, "%s.conflict", File_ID);
            ndb_set_string_to_lower(filename);
            fp = fopen(filename, "w");
            fprintf(fp, "%s\n", _conflict.Text());
            fclose (fp);
       }
       if (_special && _n_special && _generate_flag) {
            sprintf(filename, "%s.special", File_ID);
            ndb_set_string_to_lower(filename);
            fp = fopen(filename, "w");
            int num = 0, atomSerialNo = 0, WaterSerialNo = 0;
            char FieldInfo[MxField][MxLen];
            for (int i = 0; i < _n_special; i++) {
                 _ATOM **atom = _molecules[0]->find_ndb_atom(_special[i]->a_atomtyp,
                           _special[i]->a_chnid, _special[i]->a_resnam,
                           _special[i]->a_resnum, "", num);
                 if (!atom) continue;
                 if (atom[0]->occ() < 0.99) continue;

                 strcpy(FieldInfo[0], atom[0]->type());
                 sprintf(FieldInfo[1], "%d", atom[0]->atnum());
                 atomSerialNo = atom[0]->atnum() - 1;
                 strcpy(FieldInfo[2], atom[0]->atmtype());
                 strcpy(FieldInfo[3], atom[0]->alt_loc());
                 strcpy(FieldInfo[4], atom[0]->restype());
                 strcpy(FieldInfo[5], atom[0]->chnid());
                 sprintf(FieldInfo[6], "%d", atom[0]->resnum());
                 strcpy(FieldInfo[7], atom[0]->ins_code());
                 sprintf(FieldInfo[8], "%.4f", atom[0]->orig().x);
                 sprintf(FieldInfo[9], "%.4f", atom[0]->orig().y);
                 sprintf(FieldInfo[10], "%.4f", atom[0]->orig().z);
                 sprintf(FieldInfo[11], "%6.2f", atom[0]->occ());
                 sprintf(FieldInfo[12], "%6.2f", atom[0]->t_fct());
                 if (atom[0]->footnote() != -10000)
                      sprintf(FieldInfo[13], "%d", atom[0]->footnote());
                 else strcpy(FieldInfo[13], "");
                 strcpy(FieldInfo[14], atom[0]->not_in_asym());
                 sprintf(FieldInfo[15], "%d", atom[0]->pdb_resnum());
                 strcpy(FieldInfo[16], atom[0]->pdb_resnam());
                 strcpy(FieldInfo[17], atom[0]->pdb_chnid());
                 strcpy(FieldInfo[18], atom[0]->pdb_atmnam());
                 sprintf(FieldInfo[19], "%d", atom[0]->pub_resnum());
                 strcpy(FieldInfo[20], atom[0]->pub_resnam());
                 strcpy(FieldInfo[21], atom[0]->pub_chnid());
                 strcpy(FieldInfo[22], atom[0]->pub_atmnam());
                 strcpy(FieldInfo[23], atom[0]->atom_type());
                 print_atom_card(fp, HETAS_TOKEN-1, HETAS_TOKEN-1, atomSerialNo,
                   WaterSerialNo, FieldInfo, NDB_FILE_FORMAT_PDB, atom[0]->align(), 0);
                 delete [] atom;
            }
            fclose (fp);
       }
}

void _VALIDATION::finding_missing_and_extra_atoms(const int exclude_hydrogen)
{
       int need_hydrogen = 0;
       if (_experiment_type & EXPERIMENT_TYPE_NMR ||
           _experiment_type & EXPERIMENT_TYPE_NMR_SOLID) need_hydrogen = 1;
       if (exclude_hydrogen) need_hydrogen = 0;

       CifString cs;
       TblIndexObj EmptyAtoms, AA_N_Terminal_Allowed_Atoms;
       TblIndexObj AA_N_Terminal_Leaving_Atoms, AA_N_Terminal_Not_Allowed_Atoms;
       TblIndexObj AA_Leaving_Atoms, AA_Not_Allowed_Atoms, NA_5_Terminal_Allowed_Atoms;
       TblIndexObj NA_5_Terminal_Leaving_Atoms, NA_5_Terminal_Not_Allowed_Atoms;
       TblIndexObj NA_C_Terminal_Leaving_Atoms, NA_Leaving_Atoms, NA_Not_Allowed_Atoms;
       TblIndexObj ASP_N_Terminal_Leaving_Atoms, ASP_Leaving_Atoms;
       TblIndexObj CYS_N_Terminal_Leaving_Atoms, CYS_Leaving_Atoms;
       TblIndexObj GLU_N_Terminal_Leaving_Atoms, GLU_Leaving_Atoms;
       TblIndexObj HIS_N_Terminal_Leaving_Atoms, HIS_Leaving_Atoms;
       TblIndexObj ASP_C_Terminal_Leaving_Atoms, CYS_C_Terminal_Leaving_Atoms;
       TblIndexObj GLU_C_Terminal_Leaving_Atoms, HIS_C_Terminal_Leaving_Atoms;

       EmptyAtoms.Clear();

       AA_N_Terminal_Allowed_Atoms.Clear();
       ASP_C_Terminal_Leaving_Atoms.Clear();
       CYS_C_Terminal_Leaving_Atoms.Clear();
       GLU_C_Terminal_Leaving_Atoms.Clear();
       HIS_C_Terminal_Leaving_Atoms.Clear();
       for (int i = 0; i < NUM_AA_N_TERMINAL_ATOMS; i++) {
            cs.Copy(_AA_N_Terminal_Atoms[i]);
            AA_N_Terminal_Allowed_Atoms.Add(cs);
            ASP_C_Terminal_Leaving_Atoms.Add(cs);
            CYS_C_Terminal_Leaving_Atoms.Add(cs);
            GLU_C_Terminal_Leaving_Atoms.Add(cs);
            HIS_C_Terminal_Leaving_Atoms.Add(cs);
       }
       cs.Copy("HD2");
       ASP_C_Terminal_Leaving_Atoms.Add(cs);
       cs.Copy("HG");
       CYS_C_Terminal_Leaving_Atoms.Add(cs);
       cs.Copy("HE2");
       GLU_C_Terminal_Leaving_Atoms.Add(cs);
       HIS_C_Terminal_Leaving_Atoms.Add(cs);
       cs.Copy("HD1");
       HIS_C_Terminal_Leaving_Atoms.Add(cs);

       AA_N_Terminal_Leaving_Atoms.Clear();
       ASP_N_Terminal_Leaving_Atoms.Clear();
       CYS_N_Terminal_Leaving_Atoms.Clear();
       GLU_N_Terminal_Leaving_Atoms.Clear();
       HIS_N_Terminal_Leaving_Atoms.Clear();
       for (int i = 0; i < NUM_AA_N_LEAVING_ATOMS; i++) {
            cs.Copy(_AA_N_Leaving_Atoms[i]);
            AA_N_Terminal_Leaving_Atoms.Add(cs);
            ASP_N_Terminal_Leaving_Atoms.Add(cs);
            CYS_N_Terminal_Leaving_Atoms.Add(cs);
            GLU_N_Terminal_Leaving_Atoms.Add(cs);
            HIS_N_Terminal_Leaving_Atoms.Add(cs);
       }
       cs.Copy("HD2");
       ASP_N_Terminal_Leaving_Atoms.Add(cs);
       cs.Copy("HG");
       CYS_N_Terminal_Leaving_Atoms.Add(cs);
       cs.Copy("HE2");
       GLU_N_Terminal_Leaving_Atoms.Add(cs);
       HIS_N_Terminal_Leaving_Atoms.Add(cs);
       cs.Copy("HD1");
       HIS_N_Terminal_Leaving_Atoms.Add(cs);

       AA_N_Terminal_Not_Allowed_Atoms.Clear();
       for (int i = 0; i < NUM_AA_C_TERMINAL_ATOMS; i++) {
            cs.Copy(_AA_C_Terminal_Atoms[i]);
            AA_N_Terminal_Not_Allowed_Atoms.Add(cs);
       }
 
       AA_Leaving_Atoms.Clear();
       ASP_Leaving_Atoms.Clear();
       CYS_Leaving_Atoms.Clear();
       GLU_Leaving_Atoms.Clear();
       HIS_Leaving_Atoms.Clear();
       for (int i = 0; i < NUM_AA_LEAVING_ATOMS; i++) {
            cs.Copy(_AA_Leaving_Atoms[i]);
            AA_Leaving_Atoms.Add(cs);
            ASP_Leaving_Atoms.Add(cs);
            CYS_Leaving_Atoms.Add(cs);
            GLU_Leaving_Atoms.Add(cs);
            HIS_Leaving_Atoms.Add(cs);
       }
       cs.Copy("HD2");
       ASP_Leaving_Atoms.Add(cs);
       cs.Copy("HG");
       CYS_Leaving_Atoms.Add(cs);
       cs.Copy("HE2");
       GLU_Leaving_Atoms.Add(cs);
       HIS_Leaving_Atoms.Add(cs);
       cs.Copy("HD1");
       HIS_Leaving_Atoms.Add(cs);

       AA_Not_Allowed_Atoms.Clear();
       for (int i = 0; i < NUM_AA_TERMINAL_ATOMS; i++) {
            cs.Copy(_AA_Terminal_Atoms[i]);
            AA_Not_Allowed_Atoms.Add(cs);
       }
 
       NA_5_Terminal_Allowed_Atoms.Clear();
       for (int i = 0; i < NUM_NA_5_TERMINAL_ATOMS; i++) {
            cs.Copy(_NA_5_Terminal_Atoms[i]);
            NA_5_Terminal_Allowed_Atoms.Add(cs);
       }

       NA_5_Terminal_Leaving_Atoms.Clear();
       for (int i = 0; i < NUM_NA_5_LEAVING_ATOMS; i++) {
            cs.Copy(_NA_5_Leaving_Atoms[i]);
            NA_5_Terminal_Leaving_Atoms.Add(cs);
       }

       NA_5_Terminal_Not_Allowed_Atoms.Clear();
       for (int i = 0; i < NUM_NA_3_TERMINAL_ATOMS; i++) {
            cs.Copy(_NA_3_Terminal_Atoms[i]);
            NA_5_Terminal_Not_Allowed_Atoms.Add(cs);
       }

       NA_C_Terminal_Leaving_Atoms.Clear();
       for (int i = 0; i < NUM_NA_3_LEAVING_ATOMS; i++) {
            cs.Copy(_NA_3_Leaving_Atoms[i]);
            NA_C_Terminal_Leaving_Atoms.Add(cs);
       }

       NA_Leaving_Atoms.Clear();
       for (int i = 0; i < NUM_NA_LEAVING_ATOMS; i++) {
            cs.Copy(_NA_Leaving_Atoms[i]);
            NA_Leaving_Atoms.Add(cs);
       }

       NA_Not_Allowed_Atoms.Clear();
       for (int i = 0; i < NUM_NA_TERMINAL_ATOMS; i++) {
            cs.Copy(_NA_Terminal_Atoms[i]);
            NA_Not_Allowed_Atoms.Add(cs);
       }

       for (int i = 0; i < _num_mol; i++) {
            _CHAIN *chain = _molecules[i]->GetFirstChain();
            while (chain) {
                 int is_connect = 0;
                 if ((chain->chain_type() == ATOMP_TOKEN ||
                      chain->chain_type() == ATOMN_TOKEN) &&
                      chain->SeqLen() > 1) is_connect = 1;
                 _RESIDUE *First = NULL;
                 if (chain->chain_type() == ATOMP_TOKEN) {
                      for (int j = 0; j < chain->SeqLen(); j++) {
                           if (chain->SeqRes(j)->Deleted) continue;
                           First = chain->SeqRes(j)->Residue;
                           if (First) break;
                      }
                 }
                 for (int j = 0; j < chain->SeqLen(); j++) {
                      if (chain->SeqRes(j)->Deleted) continue;
                      _RESIDUE *residue = chain->SeqRes(j)->Residue;
                      if (!residue) continue;
                      if (chain->chain_type() == ATOMP_TOKEN) {
                           _RESIDUE *Next = NULL;
                           for (int k = j + 1; k < chain->SeqLen(); k++) {
                                if (chain->SeqRes(k)->Deleted) continue;
                                Next = chain->SeqRes(k)->Residue;
                                if (Next) break;
                           }
                           if (!Next) Next = First;
                           int connect = 0;
                           if (!strcmp(residue->ResName(), "ASP") ||
                               !strcmp(residue->ResName(), "GLU"))
                                connect = is_residue_linked(residue, Next);
                           if (j == 0) {
                                if (!strcmp(residue->ResName(), "ASP"))
                                     residue->find_missing_or_extra_atoms(connect,
                                          need_hydrogen, ASP_N_Terminal_Leaving_Atoms,
                                          AA_N_Terminal_Allowed_Atoms,
                                          AA_N_Terminal_Not_Allowed_Atoms);
                                else if (!strcmp(residue->ResName(), "CYS")) 
                                     residue->find_missing_or_extra_atoms(is_connect,
                                          need_hydrogen, CYS_N_Terminal_Leaving_Atoms,
                                          AA_N_Terminal_Allowed_Atoms,
                                          AA_N_Terminal_Not_Allowed_Atoms);
                                else if (!strcmp(residue->ResName(), "GLU"))
                                     residue->find_missing_or_extra_atoms(connect,
                                          need_hydrogen, GLU_N_Terminal_Leaving_Atoms,
                                          AA_N_Terminal_Allowed_Atoms,
                                          AA_N_Terminal_Not_Allowed_Atoms);
                                else if (!strcmp(residue->ResName(), "HIS"))
                                     residue->find_missing_or_extra_atoms(is_connect,
                                          need_hydrogen, HIS_N_Terminal_Leaving_Atoms,
                                          AA_N_Terminal_Allowed_Atoms,
                                          AA_N_Terminal_Not_Allowed_Atoms);
                                else residue->find_missing_or_extra_atoms(is_connect,
                                          need_hydrogen, AA_N_Terminal_Leaving_Atoms,
                                          AA_N_Terminal_Allowed_Atoms,
                                          AA_N_Terminal_Not_Allowed_Atoms);
                           } else if (j == (chain->SeqLen() - 1)) {
                                if (!strcmp(residue->ResName(), "ASP")) 
                                     residue->find_missing_or_extra_atoms(connect,
                                          need_hydrogen, ASP_C_Terminal_Leaving_Atoms,
                                          EmptyAtoms, AA_N_Terminal_Allowed_Atoms);
                                else if (!strcmp(residue->ResName(), "CYS")) 
                                     residue->find_missing_or_extra_atoms(is_connect,
                                          need_hydrogen, CYS_C_Terminal_Leaving_Atoms,
                                          EmptyAtoms, AA_N_Terminal_Allowed_Atoms);
                                else if (!strcmp(residue->ResName(), "GLU")) 
                                     residue->find_missing_or_extra_atoms(connect,
                                          need_hydrogen, GLU_C_Terminal_Leaving_Atoms,
                                          EmptyAtoms, AA_N_Terminal_Allowed_Atoms);
                                else if (!strcmp(residue->ResName(), "HIS")) 
                                     residue->find_missing_or_extra_atoms(is_connect,
                                          need_hydrogen, HIS_C_Terminal_Leaving_Atoms,
                                          EmptyAtoms, AA_N_Terminal_Allowed_Atoms);
                                else residue->find_missing_or_extra_atoms(is_connect,
                                          need_hydrogen, AA_N_Terminal_Allowed_Atoms,
                                          EmptyAtoms, AA_N_Terminal_Allowed_Atoms);
                           } else {
                                if (!strcmp(residue->ResName(), "ASP"))
                                     residue->find_missing_or_extra_atoms(connect,
                                          need_hydrogen, ASP_Leaving_Atoms,
                                          EmptyAtoms, AA_Not_Allowed_Atoms);
                                else if (!strcmp(residue->ResName(), "CYS"))
                                     residue->find_missing_or_extra_atoms(is_connect,
                                          need_hydrogen, CYS_Leaving_Atoms,
                                          EmptyAtoms, AA_Not_Allowed_Atoms);
                                else if (!strcmp(residue->ResName(), "GLU"))
                                     residue->find_missing_or_extra_atoms(connect,
                                          need_hydrogen, GLU_Leaving_Atoms,
                                          EmptyAtoms, AA_Not_Allowed_Atoms);
                                else if (!strcmp(residue->ResName(), "HIS"))
                                     residue->find_missing_or_extra_atoms(is_connect,
                                          need_hydrogen, HIS_Leaving_Atoms,
                                          EmptyAtoms, AA_Not_Allowed_Atoms);
                                else residue->find_missing_or_extra_atoms(is_connect,
                                          need_hydrogen, AA_Leaving_Atoms,
                                          EmptyAtoms, AA_Not_Allowed_Atoms);
                           }
                      } else if (chain->chain_type() == ATOMN_TOKEN) {
                           if (j == 0)
                                residue->find_missing_or_extra_atoms(is_connect,
                                         need_hydrogen, NA_5_Terminal_Leaving_Atoms,
                                         NA_5_Terminal_Allowed_Atoms,
                                         NA_5_Terminal_Not_Allowed_Atoms);
                           else if (j == (chain->SeqLen() - 1))
                                residue->find_missing_or_extra_atoms(is_connect,
                                         need_hydrogen, NA_C_Terminal_Leaving_Atoms,
                                         EmptyAtoms, NA_5_Terminal_Allowed_Atoms);
                           else residue->find_missing_or_extra_atoms(is_connect,
                                         need_hydrogen, NA_Leaving_Atoms,
                                         EmptyAtoms, NA_Not_Allowed_Atoms);
                      } else {
                           is_connect = is_residue_linked(residue);
                           residue->find_missing_or_extra_atoms(is_connect,
                               need_hydrogen, EmptyAtoms, EmptyAtoms, EmptyAtoms);
                      }
                 }
                 chain = _molecules[i]->GetNextChain();
            }
       }
}

int _VALIDATION::write_letter_token(FILE *infile, FILE *outfile, const char *token)
{
       unsigned int pos;
       int card_id, field_no, numRemark;
       char tmp_string[MxLen];
       One_Card_Info_Format *pCard = NULL;

       for (pos = 0; pos < strlen(token); pos++) {
            if (token[pos] == ' ') break;
       }
       strncpy(tmp_string, token, pos);
       tmp_string[pos] = 0;
       card_id = find_card_id(tmp_string);
       if (card_id == -1) write_special_letter_information(infile, 
                     outfile, token);
       else {
            for (; pos < strlen(token); pos++)
            if (token[pos] != ' ') break;
            strcpy(tmp_string, &token[pos]);
            if (!strcmp(tmp_string, "")) {
                 print_general_cards(outfile, card_id, 
                          NDB_FILE_FORMAT_NDB, numRemark, 0);
                 return 1;
            } else {
                 field_no = atoi(tmp_string) -1;
                 pCard = _CardInfo[card_id].Head();
                 if (pCard && strcmp(pCard->Field(field_no), ""))
                      fprintf(outfile, "%s", pCard->Field(field_no));
                 else if (NdbToken[card_id].FieldList[field_no].FieldType 
                     == NDB_DATE_FIELD) 
                      fprintf(outfile, "%s", "yyyy-mm-dd");
            }
       }

       if (card_id == -1) return 1;
       else return 0;
}

void _VALIDATION::write_special_letter_information(FILE *infile, 
        FILE *outfile, const char *token)
{
       if (!strcmp(token, "MISSING RESIDUES"))
            ndb_write_letter_misplace_atoms(NDB_FIND_MISSING_RESIDUE, 
                     infile, outfile, token);
       else if (!strcmp(token, "MISSING ATOMS"))
            ndb_write_letter_misplace_atoms(NDB_FIND_MISSING_ATOM, 
                     infile, outfile, token);
       else if (!strcmp(token, "EXTRA ATOMS"))
            ndb_write_letter_misplace_atoms(NDB_FIND_EXTRA_ATOM, 
                     infile, outfile, token);
       else if (!strcmp(token, "HYDROGEN ATOMS")) {
            passing_missing_words(FALSE, infile, token);
       } else if (!strcmp(token, "CHIRALITY")) {
            ndb_check_stereochemistry(infile, outfile, _chirality);
       } else if (!strcmp(token, "EZ")) {
            if (_experiment_type & EXPERIMENT_TYPE_NMR ||
                _experiment_type & EXPERIMENT_TYPE_NMR_SOLID)
                 ndb_check_E_Z_labeling(infile, outfile);
            else passing_missing_words(FALSE, infile, token);
       } else if (!strcmp(token, "CLOSE CONTACT1")) 
            find_out_range_waters(outfile);
       else if (!strcmp(token, "CLOSE CONTACT2"))
            move_out_range_waters(outfile);
       else if (!strcmp(token,"STRUCTURE FACTOR"))
            write_structure_factor(infile, outfile);
       else if (!strcmp(token,"PUBLICATION"))
            write_publication_question(infile, outfile);
       else if (!strcmp(token, ".1.close"))
            output_close_contact_to_letter('a', outfile);
       else if (!strcmp(token, ".close"))
            output_close_contact_to_letter('s', outfile);
       else if (!strcmp(token, ".cov-bond-rpt"))
            ndb_find_bond_differences(outfile);
       else if (!strcmp(token, ".cov-angle-rpt"))
            ndb_find_angle_differences(outfile);
       else if (!strcmp(token, ".outside-torsions"))
            pdbx_find_outside_torsions(outfile);
       else if (!strcmp(token, ".non-cis-trans-torsions"))
            pdbx_find_non_cis_trans_torsions(outfile);
       else if (!strcmp(token, ".main-chain-planarity"))
            pdbx_find_main_chain_planarity(outfile);
       else if (!strcmp(token, ".side-chain-plane"))
            pdbx_find_side_chain_plane(outfile);
}

void _VALIDATION::find_out_range_waters(FILE *letter)
{
       int i, num = 0, atomSerialNo = 0, WaterSerialNo = 0;
       char FieldInfo[MxField][MxLen];
       char temp_line[MxLen];
       _ATOM **atom = NULL;

       if (_n_waters == 0 || _waters == NULL) {
            if (letter) fprintf(letter, "     none\n\n");
            return;
       }

       for (i = 0; i < _n_waters; i++) {
            atom = _molecules[0]->find_ndb_atom(_waters[i]->a_atomtyp,
                      _waters[i]->a_chnid, _waters[i]->a_resnam,
                      _waters[i]->a_resnum, "", num);
            if (!atom) continue;

            strcpy(FieldInfo[0], atom[0]->type());
            sprintf(FieldInfo[1], "%d", atom[0]->atnum());
            atomSerialNo = atom[0]->atnum() - 1;
            strcpy(FieldInfo[2], atom[0]->atmtype());
            strcpy(FieldInfo[3], atom[0]->alt_loc());
            strcpy(FieldInfo[4], atom[0]->restype());
            strcpy(FieldInfo[5], atom[0]->chnid());
            sprintf(FieldInfo[6], "%d", atom[0]->resnum());
            strcpy(FieldInfo[7], atom[0]->ins_code());
            sprintf(FieldInfo[8], "%.4f", atom[0]->orig().x);
            sprintf(FieldInfo[9], "%.4f", atom[0]->orig().y);
            sprintf(FieldInfo[10], "%.4f", atom[0]->orig().z);
            sprintf(FieldInfo[11], "%6.2f", atom[0]->occ());
            sprintf(FieldInfo[12], "%6.2f", atom[0]->t_fct());
            if (atom[0]->footnote() != -10000)
                 sprintf(FieldInfo[13], "%d", atom[0]->footnote());
            else strcpy(FieldInfo[13], "");
            strcpy(FieldInfo[14], atom[0]->not_in_asym());
            sprintf(FieldInfo[15], "%d", atom[0]->pdb_resnum());
            strcpy(FieldInfo[16], atom[0]->pdb_resnam());
            strcpy(FieldInfo[17], atom[0]->pdb_chnid());
            strcpy(FieldInfo[18], atom[0]->pdb_atmnam());
            sprintf(FieldInfo[19], "%d", atom[0]->pub_resnum());
            strcpy(FieldInfo[20], atom[0]->pub_resnam());
            strcpy(FieldInfo[21], atom[0]->pub_chnid());
            strcpy(FieldInfo[22], atom[0]->pub_atmnam());
            strcpy(FieldInfo[23], atom[0]->atom_type());
            print_atom_card(temp_line, HETAS_TOKEN-1, HETAS_TOKEN-1, atomSerialNo,
              WaterSerialNo, FieldInfo, NDB_FILE_FORMAT_PDB, atom[0]->align(), 0);
            float dist = cal_min_distance(_molecules[0], atom[0], WATER_WATER_LIMIT);
            fprintf(letter, "%s      DIST = %6.2f A\n", temp_line, dist);
            if (dist > 50) {
                 // fprintf(letter, "\nWARNING: Water is %.3f A away.\n\n", dist);
                 _Error_Check[73]++;
            }
            delete [] atom;
       }
}

void _VALIDATION::move_out_range_waters(FILE *letter)
{
       int i, num = 0, atomSerialNo = 0, WaterSerialNo = 0;
       char FieldInfo[MxField][MxLen];
       COORD coord;
       _ATOM **atom = NULL;

       return;

       if (_n_moving_waters == 0 || _moving_waters == NULL) {
            if (letter) fprintf(letter, "     none\n\n");
            return;
       }

       for (i = 0; i < _n_moving_waters; i++) {
            atom = _molecules[0]->find_ndb_atom(_moving_waters[i]->a_atomtyp,
                      _moving_waters[i]->a_chnid, _moving_waters[i]->a_resnam,
                      _moving_waters[i]->a_resnum, "", num);
            if (!atom) continue;

            coord = atom[0]->orig();
            ndb_trans_coord_3by3(coord, _cell, NDB_TRANS_ORTHO_TO_FRAC);
            do_symmetry_operation(coord, coord, &(_cell->symops[_moving_waters[i]->sym]));
            coord.x += (float) _moving_waters[i]->transl[0];
            coord.y += (float) _moving_waters[i]->transl[1];
            coord.z += (float) _moving_waters[i]->transl[2];
            ndb_trans_coord_3by3(coord, _cell, NDB_TRANS_FRAC_TO_ORTHO);

            strcpy(FieldInfo[0], atom[0]->type());
            sprintf(FieldInfo[1], "%d", atom[0]->atnum());
            atomSerialNo = atom[0]->atnum() - 1;
            strcpy(FieldInfo[2], atom[0]->atmtype());
            strcpy(FieldInfo[3], atom[0]->alt_loc());
            strcpy(FieldInfo[4], atom[0]->restype());
            strcpy(FieldInfo[5], atom[0]->chnid());
            sprintf(FieldInfo[6], "%d", atom[0]->resnum());
            strcpy(FieldInfo[7], atom[0]->ins_code());
            sprintf(FieldInfo[8], "%.4f", coord.x);
            sprintf(FieldInfo[9], "%.4f", coord.y);
            sprintf(FieldInfo[10], "%.4f", coord.z);
            sprintf(FieldInfo[11], "%6.2f", atom[0]->occ());
            sprintf(FieldInfo[12], "%6.2f", atom[0]->t_fct());
            if (atom[0]->footnote() != -10000)
                 sprintf(FieldInfo[13], "%d", atom[0]->footnote());
            else strcpy(FieldInfo[13], "");
            strcpy(FieldInfo[14], atom[0]->not_in_asym());
            sprintf(FieldInfo[15], "%d", atom[0]->pdb_resnum());
            strcpy(FieldInfo[16], atom[0]->pdb_resnam());
            strcpy(FieldInfo[17], atom[0]->pdb_chnid());
            strcpy(FieldInfo[18], atom[0]->pdb_atmnam());
            sprintf(FieldInfo[19], "%d", atom[0]->pub_resnum());
            strcpy(FieldInfo[20], atom[0]->pub_resnam());
            strcpy(FieldInfo[21], atom[0]->pub_chnid());
            strcpy(FieldInfo[22], atom[0]->pub_atmnam());
            strcpy(FieldInfo[23], atom[0]->atom_type());
            print_atom_card(letter, HETAS_TOKEN-1, HETAS_TOKEN-1, atomSerialNo,
              WaterSerialNo, FieldInfo, NDB_FILE_FORMAT_PDB, atom[0]->align(), 0);
            delete [] atom;
       }
}

void _VALIDATION::output_close_contact_to_letter(const char type, FILE *letter)
{
       int i, count, count1 = 0;
       if (type == 'a') {
            if (_n_a_contact && _a_contact) { 
                 count = 0;
                 for (i = 0; i < _n_a_contact; i++) {
                      /* if (skip_contact(_a_contact[i])) count++;
                      else */ if (find_a_link(_a_contact[i])) count++;
                 }
                 if (count < _n_a_contact) {
                      fprintf(letter, "   Chain Atom    Res  Seq   Chain Atom    Res  Seq");
/*
                      if (_experiment_type & EXPERIMENT_TYPE_NMR ||
                          _experiment_type & EXPERIMENT_TYPE_NMR_SOLID)
*/
                      if (_cell->is_artifical)
                           fprintf(letter, "   Mol_ID            Distance\n");
                      else fprintf(letter, "   Symm_Code         Distance\n");
                      fprintf(letter, "   -----------------------------------------");
                      fprintf(letter, "--------------------------------\n");
                      if (_generate_flag) {
                           _closeHeader.Copy("   Chain Atom    Res  Seq   Chain Atom    Res  Seq");
                           if (_cell->is_artifical)
                                _closeHeader += "   Mol_ID            Distance\n";
                           else _closeHeader += "   Symm_Code         Distance\n";
                      }
                      output_close_contact(_a_contact, _n_a_contact, 1, letter, count1, -1);
                 } else fprintf(letter, "      none\n\n");
            } else fprintf(letter, "      none\n\n");
       } else if (type == 's') {
            if (_n_s_contact && _s_contact) {
/*
                 count = 0;
                 for (i = 0; i < _n_s_contact; i++) {
                      if (skip_contact(_s_contact[i])) count++;
                      else if (find_a_link(_s_contact[i])) count++;
                 }
                 if (count < _n_s_contact) {
*/
                      fprintf(letter, "   Chain Atom    Res  Seq   Chain Atom    Res  Seq");
                      fprintf(letter, "   Symm_Code         Distance\n");
                      fprintf(letter, "   -----------------------------------------");
                      fprintf(letter, "--------------------------------\n");
                      if (_generate_flag) {
                           _closeHeader.Copy("   Chain Atom    Res  Seq   Chain Atom    Res  Seq");
                           _closeHeader += "   Symm_Code         Distance\n";
                      }
                      output_close_contact(_s_contact, _n_s_contact, 1, letter, count1, -1);
                 // } else fprintf(letter, "      none\n\n");
            } else fprintf(letter, "      none\n\n");
       }
}

void _VALIDATION::output_close_contact(CONTACT **contact, const int ncontact,
            const int skip, FILE *outfile, int &count, const int max_count)
{
       int num = 0, i;
       char buffer[100], atomName_a[5], atomName_b[5], alt_a[4], alt_b[4];
       _ATOM **atom_a = NULL, **atom_b = NULL;
       CifString cs;

       for (i = 0; i < ncontact; i++) {
            if (skip && /* (skip_contact(contact[i]) || */ find_a_link(contact[i] /* ) */ )) continue;

            atom_a = _molecules[0]->find_ndb_atom(contact[i]->a_atomtyp,
                      contact[i]->a_chnid, contact[i]->a_resnam,
                      contact[i]->a_resnum, "", num);
            atom_b = _molecules[0]->find_ndb_atom(contact[i]->b_atomtyp,
                      contact[i]->b_chnid, contact[i]->b_resnam,
                      contact[i]->b_resnum, "", num);
            if (!atom_a || !atom_b) {
                 if (atom_a) delete [] atom_a;
                 if (atom_b) delete [] atom_b;
                 continue;
            }
            strcpy(alt_a, "   ");
            if (contact[i]->a_alt_loc != '\0') sprintf(alt_a, "(%c)", contact[i]->a_alt_loc);
            strcpy(alt_b, "   ");
            if (contact[i]->b_alt_loc != '\0') sprintf(alt_b, "(%c)", contact[i]->b_alt_loc);

            print_atom_atomName_field(atom_a[0]->atom_type(), 0, atomName_a,
               atom_a[0]->pdb_atmnam(), atom_a[0]->pdb_resnam(), atom_a[0]->align());
            print_atom_atomName_field(atom_b[0]->atom_type(), 0, atomName_b,
               atom_b[0]->pdb_atmnam(), atom_b[0]->pdb_resnam(), atom_b[0]->align());

            sprintf(buffer, "   %5s %s%s %3s %4d - ", atom_a[0]->pdb_chnid(),
                atomName_a, alt_a, atom_a[0]->pdb_resnam(), atom_a[0]->pdb_resnum());
            cs.Copy(buffer);
            fprintf(outfile, "%s", buffer);
            sprintf(buffer, "%5s %s%s %3s %4d", atom_b[0]->pdb_chnid(),
                atomName_b, alt_b, atom_b[0]->pdb_resnam(), atom_b[0]->pdb_resnum());
            cs += buffer;
            fprintf(outfile, "%s", buffer);
/*
            if (_experiment_type & EXPERIMENT_TYPE_NMR ||
                _experiment_type & EXPERIMENT_TYPE_NMR_SOLID)
*/
            if (_cell->is_artifical)
                 sprintf(buffer, "    %5d            Dist = %.2lf\n",
                     contact[i]->Mol_ID, contact[i]->dist);
            else sprintf(buffer, "  ( %2d, %2d, %2d, %2d)  Dist = %.2lf\n",
                     contact[i]->sym, contact[i]->transl[0]+5,
                     contact[i]->transl[1]+5, contact[i]->transl[2]+5,
                     contact[i]->dist);
            cs += buffer;
            fprintf(outfile, "%s", buffer);
            _Error_Check[0]++;
            if (atom_a) delete [] atom_a;
            if (atom_b) delete [] atom_b;

            if (_generate_flag && !skip_contact(contact[i]) &&
                contact[i]->dist < CLOSE_CUTOFF) {
                 if (_closeHeader != "") {
                      _closeContact += _closeHeader;
                      _closeHeader.Clear();
                 }
                 _closeContact += cs;
            }

            count++;
            if (max_count > 0 && count > max_count) break;
       }
}

void _VALIDATION::ndb_find_bond_differences(FILE *fp)
{
       int first = 1, i = 0, j = 0, len;
       double hi = 0, low = 0, d = 0, range = 0, d2 = 0.0, rmsbond = 0.0;
       char buffer[100], chnid[4], Fst_Atom[5], Snd_Atom[5], *p = NULL;

       for (i = 0; i < _n_bond; i++) {
            d = _bond[i]->val - _bond[i]->ept;
            d2 += d * d;
       }
       if (_n_bond) rmsbond = sqrt ((d2) / (double) _n_bond);

#ifdef EH99
       fprintf(fp, "    The overall RMS deviation for covalent bonds relative to the standard \n");
#else
       fprintf(fp, "    The RMS deviation for covalent bonds relative to the standard \n");
#endif
       fprintf(fp, "    dictionary is %7.3f Angstroms\n", rmsbond);

       range =  6.0 * rmsbond;
       CifString Header, cs;
       Header.Copy("\nCovalent bond lengths:\n");
       if (_num_mol > 1)
            Header += " Deviation  Residue  Chain  Sequence  Model  ";
       else Header += " Deviation  Residue  Chain  Sequence  ";
       Header += "AT1  -  AT2    Bond    Dictionary  Standard\n";
       Header += "             Name     ID     Number   ";
       Header += "              Distance    Value    Deviation\n";

       int count_total = 0;
       int error_total = 0;
       for (i = 0; i < _n_bond; i++) {
            if ((strcmp(_bond[i]->fst->alt_loc(), "") && strcmp(_bond[i]->fst->alt_loc(), "A")) ||
                (strcmp(_bond[i]->snd->alt_loc(), "") && strcmp(_bond[i]->snd->alt_loc(), "A"))) continue;

            count_total++;
#ifdef EH99
            range = SIGMA4_6 * _bond[i]->std;
#endif
            hi  = _bond[i]->ept + range;
            low = _bond[i]->ept - range;
            if (_bond[i]->val < low || _bond[i]->val > hi) {
                 error_total++;
                 d = _bond[i]->val - _bond[i]->ept;
                 if (first) {
                      first = 0;
                      fprintf(fp, "\n\n    The following table contains a list ");
                      fprintf(fp, "of the covalent bonds\n");
#ifdef EH99
                      fprintf(fp, "    greater than %d times standard deviation.\n\n\n", SIGMA4_6);
#else
                      fprintf(fp, "    greater than 6.0*RMSD.\n\n\n");
#endif
                      if (_num_mol > 1)
                           fprintf(fp, " Deviation  Residue  Chain  Sequence  Model  ");
                      else fprintf(fp, " Deviation  Residue  Chain  Sequence  ");
                      fprintf(fp, "AT1  -  AT2    Bond    Dictionary  Standard\n");
                      fprintf(fp, "             Name     ID     Number   ");
                      fprintf(fp, "              Distance    Value    Deviation\n");
                      fprintf(fp, "--------------------------------------");
                      fprintf(fp, "--------------------------------------------\n");
                 }

                 if (strcmp(_bond[i]->ChainID, ""))
                      strcpy(chnid, _bond[i]->ChainID);
                 else strcpy(chnid, " ");
                 strcpy(buffer, _bond[i]->type);
                 len = strlen(buffer);
                 for (j = 0; j < len; j++) {
                      if (buffer[j] == '-') buffer[j] = ' ';
                 }
                 p = get_word(buffer, Fst_Atom);
                 p = get_word(p, Snd_Atom);

                 if (_num_mol > 1)
                      sprintf(buffer, " %7.3f    %4s        %s   %4d      %3d   ",
                           d, _bond[i]->ResName, chnid, _bond[i]->ResNum, _bond[i]->Mol_ID);
                 else sprintf(buffer, " %7.3f    %4s        %s   %4d      ",
                           d, _bond[i]->ResName, chnid, _bond[i]->ResNum);
                 fprintf(fp, "%s", buffer);
                 cs.Copy(buffer);
                 sprintf(buffer, "%-4s -  %-4s %7.3f   %7.3f     %7.3f\n",
                           Fst_Atom, Snd_Atom, _bond[i]->val, _bond[i]->ept, _bond[i]->std);
                 fprintf(fp, "%s", buffer);
                 cs += buffer;
                 if (_generate_flag && fabs(d) > BOND_CUTOFF) {
                      if (Header != "") {
                           _bond_angle_torsion += Header;
                           Header.Clear();
                      }
                      _bond_angle_torsion += cs;
                 }
                 _Error_Check[1]++;
            }
       }

       if (first) {
            fprintf(fp, "\n    All covalent bonds lie within a 6.0*RMSD range about the \n");
            fprintf(fp, "    standard dictionary values.\n");
       } else {
            fprintf(fp, "\n\n    %d out of total %d covalent bonds (%.3f%%) ", error_total,
                        count_total, (double) error_total * 100 / (double) count_total);
            if (count_total > 1)
                 fprintf(fp, "have ");
            else fprintf(fp, "has ");
            fprintf(fp, "greater than %d times\n    standard deviation.\n", SIGMA4_6);
       }
}

void _VALIDATION::ndb_find_angle_differences(FILE *fp)
{
       int first = 1, i = 0, j = 0, len;
       double hi = 0, low = 0, d = 0, d2 = 0.0, rmsangle = 0.0, range = 0;
       char buffer[100], chnid[4], Fst_Atom[5], Snd_Atom[5], Trd_Atom[5], *p = NULL;

       for (i = 0; i < _n_angle; i++) {
            d = _angle[i]->val - _angle[i]->ept;
            d2 += d * d;
       }
       if (_n_angle) rmsangle = sqrt ((d2) / (double) _n_angle);

#ifdef EH99
       fprintf(fp, "    The overall RMS deviation for covalent angles relative to the standard \n");
#else
       fprintf(fp, "    The RMS deviation for covalent angles relative to the standard \n");
#endif
       fprintf(fp, "    dictionary is %6.1f degrees.\n",rmsangle);

       range = 6.0 * rmsangle;
       CifString Header, cs;
       Header.Copy("\nCovalent bond angles:\n");
       if (_num_mol > 1)
            Header += " Deviation  Residue  Chain  Sequence  Model  ";
       else Header += " Deviation  Residue  Chain  Sequence  ";
       Header += "AT1  -  AT2  -  AT3    Bond    Dictionary  Standard\n";
       Header += "             Name     ID     Number   ";
       if (_num_mol > 1)
            Header += "                              Angle      Value    Deviation\n";
       else Header += "                       Angle      Value    Deviation\n"; 

       int count_total = 0;
       int error_total = 0;
       for (i = 0; i < _n_angle; i++) {
            if ((strcmp(_angle[i]->fst->alt_loc(), "") && strcmp(_angle[i]->fst->alt_loc(), "A")) ||
                (strcmp(_angle[i]->snd->alt_loc(), "") && strcmp(_angle[i]->snd->alt_loc(), "A")) ||
                (strcmp(_angle[i]->trd->alt_loc(), "") && strcmp(_angle[i]->trd->alt_loc(), "A"))) continue;

            count_total++;
#ifdef EH99
            range = SIGMA4_6 * _angle[i]->std;
#endif
            hi  = _angle[i]->ept + range;
            low = _angle[i]->ept - range;
            if (_angle[i]->val < low || _angle[i]->val > hi) {
                 error_total++;
                 d = _angle[i]->val - _angle[i]->ept;
                 if (first) {
                      first = 0;
                      fprintf(fp, "\n    The following table contains a list ");
                      fprintf(fp, "of the covalent bond angles\n");
#ifdef EH99
                      fprintf(fp, "    greater than %d times standard deviation.\n\n\n", SIGMA4_6);
#else
                      fprintf(fp, "    greater than 6.0*RMSD.\n\n\n");
#endif
                      if (_num_mol > 1) {
                           fprintf(fp, " Deviation  Residue  Chain  Sequence  Model  ");
                           fprintf(fp, "AT1  -  AT2  -  AT3    Bond    Dictionary  Standard\n");
                           fprintf(fp, "             Name     ID     Number   ");
                           fprintf(fp, "                              Angle      Value    Deviation\n");
                           fprintf(fp, "---------------------------------------");
                           fprintf(fp, "-----------------------------------------------------------\n");
                      } else {
                           fprintf(fp, " Deviation  Residue  Chain  Sequence  ");
                           fprintf(fp, "AT1  -  AT2  -  AT3    Bond    Dictionary  Standard\n");
                           fprintf(fp, "             Name     ID     Number   ");
                           fprintf(fp, "                       Angle      Value    Deviation\n");
                           fprintf(fp, "---------------------------------------");
                           fprintf(fp, "----------------------------------------------------\n");
                      }
                 }

                 if (strcmp(_angle[i]->ChainID, ""))
                      strcpy(chnid, _angle[i]->ChainID);
                 else strcpy(chnid, " ");
                 strcpy(buffer, _angle[i]->type);
                 len = strlen(buffer);
                 for (j = 0; j < len; j++) {
                      if (buffer[j] == '-') buffer[j] = ' ';
                 }
                 p = get_word(buffer, Fst_Atom);
                 p = get_word(p, Snd_Atom);
                 p = get_word(p, Trd_Atom);

                 if (_num_mol > 1)
                      sprintf(buffer, " %7.1f   %4s       %s     %4d       %3d   ",
                           d, _angle[i]->ResName, chnid, _angle[i]->ResNum, _angle[i]->Mol_ID);
                 else sprintf(buffer, " %7.1f   %4s       %s     %4d      ",
                           d, _angle[i]->ResName, chnid, _angle[i]->ResNum);
                 fprintf(fp, "%s", buffer);
                 cs.Copy(buffer);
                 sprintf(buffer, "%-4s -  %-4s -  %-4s %7.1f   %7.1f    %7.1f\n",
                           Fst_Atom, Snd_Atom, Trd_Atom, _angle[i]->val, _angle[i]->ept,
                           _angle[i]->std);
                 fprintf(fp, "%s", buffer);
                 cs += buffer;
                 if (_generate_flag && fabs(d) > ANGLE_CUTOFF) {
                      if (Header != "") {
                           _bond_angle_torsion += Header;
                           Header.Clear();
                      }
                      _bond_angle_torsion += cs;
                 }
                 _Error_Check[2]++;
            }
       }

       if (first) {
            fprintf(fp, "\n\n    All covalent bond angles lie within a 6.0*RMSD range about the \n");
            fprintf(fp, "    standard dictionary values.\n");
       } else {
            fprintf(fp, "\n\n    %d out of total %d bond angles (%.3f%%) ", error_total,
                        count_total, (double) error_total * 100 / (double) count_total);
            if (count_total > 1)
                 fprintf(fp, "have ");
            else fprintf(fp, "has ");
            fprintf(fp, "greater than %d times\n    standard deviation.\n", SIGMA4_6);
       }
}

void _VALIDATION::write_structure_factor(FILE *infile, FILE *outfile) 
{
       int foundFlag = FALSE, ret, i, len;
       char line[MxLen], token[MxLen];
       One_Card_Info_Format *pCard = _CardInfo[RECVD1_TOKEN-1].Head();
       while (pCard) {
            if (!strcmp(pCard->Field(5), "Y")) break;
            pCard = pCard->Next();
       }
       if (!pCard)  {
            passing_missing_words(FALSE, infile, "STRUCTURE FACTOR");
            return;
       } else {
            while (!feof(infile)) {
                 if (fgets(line,MxLen,infile));
                 len = strlen(line);
                 for (i = 0; i < len; i++)  {
                      ret = find_letter_token(&line[i], token);
                      if (ret == FALSE) fprintf(outfile, "%c", line[i]);
                      else {
                           i += ret;
                           if (strcmp(token, "STRUCTURE FACTOR")) 
                                write_letter_token(infile, outfile, token);
                           else {
                                foundFlag = TRUE; break;
                           }
                      }
                 }
                 if (foundFlag) break;
            }
       }
}

void _VALIDATION::write_publication_question(FILE *infile, FILE *outfile) 
{
       int i, ret, ifFlag = FALSE, len;
       char tmp_string[MxLen], line[MxLen], token[MxLen];
       One_Card_Info_Format *pCard = NULL;

       for (i = 0; i < MxJrnlTokens; i++)
            if (!strcmp(JrnlToken[i].TokenName, "REF")) break;
       pCard = _JrnlCardInfo[i].Head();
       if (pCard && !strcmp(pCard->Field(1), "")) {
            strcpy(tmp_string, pCard->Field(4));
            ndb_set_string_to_upper(tmp_string);
            if (!strncmp(tmp_string, "TO BE PUBLISHED", 15))
                 ifFlag = TRUE;
       } else ifFlag = TRUE;

       if (!ifFlag) passing_missing_words(FALSE, infile, "else");
       while (!feof(infile)) {
            if (fgets(line, MxLen, infile));
            len = strlen(line);
            for (i = 0; i < len; i++)  {
                 ret = find_letter_token(&line[i], token);
                 if (ret == FALSE) fprintf(outfile, "%c", line[i]);
                 else {
                      i += ret;
                      if (strcmp(token, "PUBLICATION") && strcmp(token, "else"))
                           write_letter_token(infile, outfile, token);
                      else break;
                 }
            }
            if (i != len) break;
       }
       if (ifFlag) passing_missing_words(FALSE, infile, "PUBLICATION");
}

void _VALIDATION::ndb_write_letter_misplace_atoms(const int flag, FILE *infile,
            FILE *outfile, const char *token)
{
       int first = TRUE, i = 0, j, pos = 0, need_hydrogen = 0;
       char Chain_ID[5], Mol_ID[4], insert_code[2], buff[10], *p = NULL, line[200];
       _CHAIN *chain = NULL;
       _RESIDUE *residue = NULL;
       One_Card_Info_Format *pCard = NULL;

       if (flag == NDB_FIND_MISSING_RESIDUE) {
            pCard = _CardInfo[MISRES_TOKEN-1].Head();
            if (pCard) {
                 ndb_writing_missing_words(infile, outfile);
                 first = FALSE;
            }
            while (pCard) {
                 strcpy(Chain_ID, " ");
                 strcpy(Mol_ID, "   ");
                 strcpy(insert_code, " ");
                 if (strcmp(pCard->Field(2), "")) strcpy(Mol_ID, pCard->Field(2));
                 if (strcmp(pCard->Field(3), "")) strcpy(Chain_ID, pCard->Field(3));
                 if (strcmp(pCard->Field(6), "")) strcpy(insert_code, pCard->Field(6));
                 fprintf(outfile, " %7s(%3s %s%4s%c)\n", pCard->Field(4), Mol_ID,
                              Chain_ID, pCard->Field(5), insert_code[0]);
                 pCard = pCard->Next();
            }
       } else {
            if (_experiment_type & EXPERIMENT_TYPE_NMR ||
                _experiment_type & EXPERIMENT_TYPE_NMR_SOLID) need_hydrogen = 1;
            for (i = 0; i < _num_mol; i++) {
                 chain = _molecules[i]->GetFirstChain();
                 while (chain) {
                      if (flag == NDB_FIND_EXTRA_ATOM && need_hydrogen &&
                          chain->chain_type() == ATOMP_TOKEN &&
                          !chain->SeqRes(0)->Deleted) {
                           residue = chain->SeqRes(0)->Residue;
                           if (residue) {
                                for (j = 0; j < _n_link; j++) {
                                     if ((_link[j]->FstIndex == chain->index() &&
                                         _link[j]->FstRes == 0 &&
                                         !strcmp(_link[j]->FstAtom, "N")) ||
                                        (_link[j]->SndIndex == chain->index() &&
                                         _link[j]->SndRes == 0 &&
                                         !strcmp(_link[j]->SndAtom, "N")))
                                          break;
                                }
                                if (j == _n_link) residue->find_extra_H_atom();
                           }
                      }
                      residue = chain->GetFirstResidue();
                      while (residue) {
                           strcpy(Chain_ID, " ");
                           if (strcmp(chain->PDB_ChainID(), ""))
                                strcpy(Chain_ID, chain->PDB_ChainID());

                           if (flag == NDB_FIND_MISSING_ATOM)
                                p = residue->missing().Text();
                           else if (flag == NDB_FIND_EXTRA_ATOM)
                                p = residue->extras().Text();
                           if (p && strcmp(p, "")) {
                                if (flag == NDB_FIND_EXTRA_ATOM) _Error_Check[21]++;
                                ndb_clean_string(p);
                                      if (first) {
                                     ndb_writing_missing_words(infile, outfile);
                                     first = FALSE;
                                }
                                if (_num_mol > 1)
                                     fprintf(outfile, " %7s(%3d %c%4d)        ",
                                         residue->ResName(), _molecules[i]->Mol_ID(),
                                         Chain_ID[0], residue->pdb_res_no());
                                else fprintf(outfile, " %7s(    %c%4d)        ",
                                         residue->ResName(), Chain_ID[0],
                                         residue->pdb_res_no());
                                pos = 27;
                                while (p) {
                                     p = get_word(p, buff);
                                     pos += 5;
                                     if (pos > 81) {
                                          fprintf(outfile,
                                              "\n                           ");
                                          pos = 27;
                                     }
                                     fprintf(outfile, "%4s ", buff);
                                }
                                fprintf(outfile, "\n");
                           }
                           residue = chain->GetNextResidue();
                      }
                      chain = _molecules[i]->GetNextChain();
                 }
            }
       }
       if (first) passing_missing_words(FALSE, infile, token);

       if (flag == NDB_FIND_MISSING_RESIDUE) {
            fputc('\n', outfile);
            FILE *fp = fopen("validation.alignment", "r");
            if (fp) {
                  while (!feof(fp)) {
                       if (fgets(line, 199, fp));
                       if (feof(fp)) break;
                       fprintf(outfile, "%s", line);
                       if (!strncmp(line, "SEQUENCE WARNING: ", 18))
                            _backbone += &line[18];
                       // else if (!strncmp(line, " C-N bond distance is", 21))
                       else if (strstr(line, "bond distance is"))
                            _backbone += line;
                       else if (!strncmp(line, "   MISMATCH: ", 13)) {
                            if (!strstr(line, "ALA)") && !strstr(line, "GLY)"))
                                 _conflict += line;
                       }
                  }
                  fclose (fp);
            }
       }
}

void _VALIDATION::check_link_records(FILE *fp)
{
       int first = 1, found = 0, i = 0, a, b, num_a = 0, num_b = 0;
       char chnid1[5], chnid2[5], loc1[5], loc2[5], buffer[10];
       _ATOM **atom_a = NULL, **atom_b = NULL;
       _CHAIN *ch1 = NULL, *ch2 = NULL;
       _FIELD *Init = NULL, *End = NULL;
       _MOLECULE *mol = NULL;
       _RESIDUE *beg = NULL, *end = NULL;
       float dist = 0;

       if (_num_mol == 0 || _molecules == NULL) return;

       std::set<std::string> link_set;
       link_set.clear();
       for (i = 0; i < _n_link; i++) {
            mol = _molecules[_link[i]->mol_index];
            ch1 = mol->GetIndexChain(_link[i]->FstIndex);
            ch2 = mol->GetIndexChain(_link[i]->SndIndex);
            if (!ch1 || !ch2) continue;
            Init = ch1->SeqResIndex(_link[i]->FstRes);
            End  = ch2->SeqResIndex(_link[i]->SndRes);
            if (!Init || !End) continue;
            beg = Init->Residue;
            end = End->Residue;
            if (!beg || !end) continue;

            found = 0;
            atom_a = beg->find_atom(_link[i]->FstAtom, num_a);
            atom_b = end->find_atom(_link[i]->SndAtom, num_b);
            if (atom_a && atom_b) {
                 std::string link_a = atom_a[0]->pdb_chnid();
                 link_a += "_"; link_a += atom_a[0]->pdb_resnam();
                 sprintf(buffer, "%d", atom_a[0]->pdb_resnum());
                 link_a += "_"; link_a += buffer;
                 link_a += "_"; link_a += atom_a[0]->ins_code();
                 std::string link_b = atom_b[0]->pdb_chnid();
                 link_b += "_"; link_b += atom_b[0]->pdb_resnam();
                 sprintf(buffer, "%d", atom_b[0]->pdb_resnum());
                 link_b += "_"; link_b += buffer;
                 link_b += "_"; link_b += atom_b[0]->ins_code();
                 link_set.insert(link_a + "_" + link_b);
                 link_set.insert(link_b + "_" + link_a);

                 found = 1;
                 for (a = 0; a < num_a; a++) {
                      if (strcmp(atom_a[a]->alt_loc(), _link[i]->altLoc1) ||
                          strcmp(atom_a[a]->restype(), _link[i]->FstName)) continue;
                      for (b = 0; b < num_b; b++) {
                           if (strcmp(atom_b[b]->alt_loc(), _link[i]->altLoc2) ||
                               strcmp(atom_b[b]->restype(), _link[i]->SndName)) continue;
                           found = 2;
                           if ((!strcmp(_link[i]->SymOP_1, "") ||
                                !strcmp(_link[i]->SymOP_1, "1_555")) &&
                               (!strcmp(_link[i]->SymOP_2, "") ||
                                !strcmp(_link[i]->SymOP_2, "1_555")))
                                dist = cal_distance(atom_a[a], atom_b[b]);
                           else if (!strcmp(_link[i]->SymOP_2, "") ||
                               !strcmp(_link[i]->SymOP_2, "1_555"))
                                dist = cal_symmetry_distance(atom_b[b]->orig(), atom_a[a]->orig(),
                                           _link[i]->SymOP_1, _cell);
                           else dist = cal_symmetry_distance(atom_a[a]->orig(), atom_b[b]->orig(),
                                            _link[i]->SymOP_2, _cell);
                           if (!is_a_link(atom_a[a]->atom_type(), atom_b[b]->atom_type(), dist)) {
                                _Error_Check[20]++;
                                sprintf(buffer, " %.3f", dist);
                                _Error_Check_Message[20] += buffer;
                                if (first) {
                                     fprintf(fp, "\n    CHECK 'LINK' RECORDS\n");
                                     fprintf(fp, "    --------------------\n\n");
                                     first = 0;
                                }
                                char atom_name_a[20], atom_name_b[20];
                                strcpy(atom_name_a, atom_a[a]->atmtype());
                                if (strcmp(atom_a[a]->alt_loc(), ""))
                                     sprintf(atom_name_a, "%s(%s)", atom_a[a]->atmtype(),
                                                atom_a[a]->alt_loc());
                                strcpy(atom_name_b, atom_b[b]->atmtype());
                                if (strcmp(atom_b[b]->alt_loc(), ""))
                                     sprintf(atom_name_b, "%s(%s)", atom_b[b]->atmtype(),
                                                atom_b[b]->alt_loc());
                                fprintf(fp, "The distance between %s %s %s %d%s",
                                      atom_name_a, atom_a[a]->pdb_resnam(),
                                      atom_a[a]->pdb_chnid(), atom_a[a]->pdb_resnum(),
                                      atom_a[a]->ins_code());
                                fprintf(fp, " and %s %s %s %d%s is %.3f A.\n",
                                      atom_name_b, atom_b[b]->pdb_resnam(),
                                      atom_b[b]->pdb_chnid(), atom_b[b]->pdb_resnum(),
                                      atom_b[b]->ins_code(), dist);
                           }
                      }
                 }
            }
            if (atom_a) delete [] atom_a;
            if (atom_b) delete [] atom_b;
            if (!found) {
                 _Error_Check[36]++;
                 if (first) {
                      fprintf(fp, "\n    CHECK 'LINK' RECORDS\n");
                      fprintf(fp, "    --------------------\n\n");
                      first = 0;
                 }
                 strcpy(chnid1, " ");
                 if (strcmp(ch1->PDB_ChainID(), "")) strcpy(chnid1, ch1->PDB_ChainID());
                 strcpy(chnid2, " ");
                 if (strcmp(ch2->PDB_ChainID(), "")) strcpy(chnid2, ch2->PDB_ChainID());
                 strcpy(loc1, " ");
                 if (strcmp(_link[i]->altLoc1, "")) strcpy(loc1, _link[i]->altLoc1);
                 strcpy(loc2, " ");
                 if (strcmp(_link[i]->altLoc2, "")) strcpy(loc2, _link[i]->altLoc2);
                 fprintf(fp, "Incorrect LINK record between (%s %d %s %s %s) --- (%s %d %s %s %s)\n",
                     beg->ResName(), beg->pdb_res_no(), chnid1, _link[i]->FstAtom, loc1,
                     end->ResName(), end->pdb_res_no(), chnid2, _link[i]->SndAtom, loc2);
            } else if (found == 1) {
                 _Error_Check[37]++;
                 if (first) {
                      fprintf(fp, "\n    CHECK 'LINK' RECORDS\n");
                      fprintf(fp, "    --------------------\n\n");
                      first = 0;
                 }
                 strcpy(chnid1, " ");
                 if (strcmp(ch1->PDB_ChainID(), "")) strcpy(chnid1, ch1->PDB_ChainID());
                 strcpy(chnid2, " ");
                 if (strcmp(ch2->PDB_ChainID(), "")) strcpy(chnid2, ch2->PDB_ChainID());
                 strcpy(loc1, " ");
                 if (strcmp(_link[i]->altLoc1, "")) strcpy(loc1, _link[i]->altLoc1);
                 strcpy(loc2, " ");
                 if (strcmp(_link[i]->altLoc2, "")) strcpy(loc2, _link[i]->altLoc2);
                 fprintf(fp, "Missing alternate conformation ID for LINK record between ");
                 fprintf(fp, "          (%s %d %s %s %s) --- (%s %d %s %s %s)\n",
                     beg->ResName(), beg->pdb_res_no(), chnid1, _link[i]->FstAtom, loc1,
                     end->ResName(), end->pdb_res_no(), chnid2, _link[i]->SndAtom, loc2);
            }
       }

       _CHAIN *chain = _molecules[0]->GetFirstChain();
       while (chain) {
            if (chain->chain_type() != ATOMP_TOKEN &&
                chain->chain_type() != ATOMN_TOKEN) {
                 chain = _molecules[0]->GetNextChain();
                 continue;
            }
            for (int j = 1; j < chain->SeqLen(); j++) {
                 if (chain->SeqRes(j-1)->Deleted ||
                     chain->SeqRes(j)->Deleted) continue;
                 beg = chain->SeqRes(j-1)->Residue;
                 end = chain->SeqRes(j)->Residue;
                 if (!beg || !end) continue;
                 if (is_a_standard_residue(beg->ResName()) &&
                     is_a_standard_residue(end->ResName())) continue;

                 _ATOM *b_first = beg->GetFirstAtom();
                 _ATOM *e_first = end->GetFirstAtom();
                 std::string link_a = b_first->pdb_chnid();
                 link_a += "_"; link_a += b_first->pdb_resnam();
                 sprintf(buffer, "%d", b_first->pdb_resnum());
                 link_a += "_"; link_a += buffer;
                 link_a += "_"; link_a += b_first->ins_code();
                 std::string link_b = e_first->pdb_chnid();
                 link_b += "_"; link_b += e_first->pdb_resnam();
                 sprintf(buffer, "%d", e_first->pdb_resnum());
                 link_b += "_"; link_b += buffer;
                 link_b += "_"; link_b += e_first->ins_code();
                 if (link_set.find(link_a + "_" + link_b) != link_set.end()) continue;

                 fprintf(fp, "Missing LINK record between (%s %d%s %s) --- (%s %d%s %s)\n",
                             b_first->pdb_resnam(), b_first->pdb_resnum(),
                             b_first->ins_code(), b_first->pdb_chnid(),
                             e_first->pdb_resnam(), e_first->pdb_resnum(),
                             e_first->ins_code(), e_first->pdb_chnid());
            }
            chain = _molecules[0]->GetNextChain();
       }
}

void _VALIDATION::check_site_records(FILE *fp)
{
       if (!_CifObj) return;
       if (_num_mol == 0 || _molecules == NULL) return;

       ISTable *t1 = _CifObj->GetTablePtr(_StructureId, "struct_site");
       if (is_empty_table(t1)) return;
       ISTable *t2 = _CifObj->GetTablePtr(_StructureId, "struct_site_gen");
       if (is_empty_table(t2)) return;

       int rowNo1 = t1->GetNumRows();
       int t1_idx1 = t1->GetColumnIndex("pdbx_evidence_code");
       int t1_idx2 = t1->GetColumnIndex("id");
       int t1_idx3 = t1->GetColumnIndex("details");
       int rowNo2 = t2->GetNumRows();
       int t2_idx1 = t2->GetColumnIndex("site_id");
       int t2_idx2 = t2->GetColumnIndex("auth_asym_id");
       int t2_idx3 = t2->GetColumnIndex("auth_comp_id");
       int t2_idx4 = t2->GetColumnIndex("auth_seq_id");
       int t2_idx5 = t2->GetColumnIndex("ndb_label_ins_code");
       int t2_idx6 = t2->GetColumnIndex("symmetry");

       ReVarCifArray<CifString> data;
       CifString cs, chn, res, ins, cs1, cs2, cs3, cs4, cs5, cs6;
       int num;
       for (int i = 0; i < rowNo1; ++i) {
            get_value(cs, t1, t1_idx1, i);
            cs.ToUpper();
            if (cs != "SOFTWARE") continue;
            get_value(cs, t1, t1_idx2, i);
            get_value(cs1, t1, t1_idx3, i);
            if (strstr(cs1.Text(), "BINDING SITE FOR CHAIN")) continue;
            if (!strstr(cs1.Text(), "BINDING SITE FOR RESIDUE")) continue;
            get_wordlist_from_string(data, cs1.Text());
            if (data.Length() < 6) {
                 _Error_Check[72]++;
                 fprintf(fp, "\nCan't parse \"%s\" for %s. ", cs1.Text(), cs.Text());
                 fprintf(fp, "Looking for 'BINDING SITE FOR RESIDUE res chn num'\n");
                 continue;
            }
            if (data.Length() == 6) {
                 res.Copy(data[4]);
                 // chn.Copy(data[5].Text()[0]);
                 chn.Clear();
                 chn += data[5].Text()[0];
                 num = atoi(&(data[5].Text()[1]));
                 ins.Clear();
                 for (unsigned int j = 1; j < strlen(data[5].Text()); ++j) {
                      if (isalpha(data[5].Text()[j])) {
                           ins.Copy(&(data[5].Text()[j]));
                           break; 
                      }
                 }
            } else {
                 res.Copy(data[4]);
                 chn.Copy(data[5]);
                 num = atoi(data[6].Text());
                 ins.Clear();
                 for (unsigned int j = 0; j < strlen(data[6].Text()); ++j) {
                      if (isalpha(data[6].Text()[j])) {
                           ins.Copy(&(data[6].Text()[j]));
                           break; 
                      }
                 }
            }
            _RESIDUE *res1 = _molecules[0]->find_pdb_residue(chn.Text(),
                                 res.Text(), num, ins.Text());
            if (!res1) {
                 _Error_Check[72]++;
                 fprintf(fp, "\nCan't find redsidue %s %s %d%s for site %s\n",
                        res.Text(), chn.Text(), num, ins.Text(), cs.Text());
                 continue;
            }

            for (int j = 0; j < rowNo2; ++j) {
                 get_value(cs1, t2, t2_idx1, j);
                 if (cs1 != cs) continue;
                 get_value(cs2, t2, t2_idx2, j);
                 get_value(cs3, t2, t2_idx3, j);
                 get_value(cs4, t2, t2_idx4, j);
                 get_value(cs5, t2, t2_idx5, j);
                 get_value(cs6, t2, t2_idx6, j);
                 _RESIDUE *res2 = _molecules[0]->find_pdb_residue(cs2.Text(),
                            cs3.Text(), atoi(cs4.Text()), cs5.Text());
                 if (!res2) {
                      _Error_Check[72]++;
                      cs4 += cs5;
                      fprintf(fp, "\nCan't find redsidue %s %s %s for site %s\n",
                             cs3.Text(), cs2.Text(), cs4.Text(), cs.Text());
                      continue;
                 }

                 bool has_contact = false;
                 _ATOM *atom1 = res1->GetFirstAtom();
                 while (atom1) {
                      _ATOM *atom2 = res2->GetFirstAtom();
                      while (atom2) {
                           if (strcmp(atom1->alt_loc(), "") &&
                               strcmp(atom2->alt_loc(), "") &&
                               strcmp(atom1->alt_loc(),
                                      atom2->alt_loc()) &&
                               strcmp(res1->ResName(), "HOH") &&
                               strcmp(res2->ResName(), "HOH")) {
                                atom2 = res2->GetNextAtom();
                                continue;
                           }
                           has_contact = true;
                           break;
                      }
                      if (has_contact) break;
                      atom1 = res1->GetNextAtom();
                 }

                 float max_dist = 10000.0;
                 float dist;
                 bool found = false;
                 atom1 = res1->GetFirstAtom();
                 while (atom1) {
                      _ATOM *atom2 = res2->GetFirstAtom();
                      while (atom2) {
                           if (cs6 != "" && cs6 != "1_555")
                                dist = cal_symmetry_distance(atom1->orig(),
                                         atom2->orig(), cs6.Text(), _cell);
                           else dist = cal_distance(atom1, atom2);
                           if (dist <= 8.00) {
                                found = true;
                                if (has_contact) break;
                           }
                           if (dist < max_dist) max_dist = dist;
                           atom2 = res2->GetNextAtom();
                      }
                      if (has_contact && found) break;
                      atom1 = res1->GetNextAtom();
                 }

                 if (has_contact && found) continue;

                 if (has_contact) {
                      _Error_Check[72]++;
                      cs4 += cs5;
                      fprintf(fp, "The closest distance between '%s %s %d%s' and '%s %s %s %s' ",
                                 res.Text(), chn.Text(), num, ins.Text(),
                                 cs3.Text(), cs2.Text(), cs4.Text(), cs6.Text());
                      fprintf(fp, "is %.2f Angstroms for site %s\n", max_dist, cs.Text());
                 } else if (max_dist < 2.0) {
                      _Error_Check[72]++;
                      cs4 += cs5;
                      fprintf(fp, "The closest distance between '%s %s %d%s' and '%s %s %s %s' ",
                                 res.Text(), chn.Text(), num, ins.Text(),
                                 cs3.Text(), cs2.Text(), cs4.Text(), cs6.Text());
                      fprintf(fp, "is %.2f Angstroms for site %s\n", 10000.0, cs.Text());
                 }
            }
       }
}

void _VALIDATION::print_link_records(const char *File_ID)
{
       if (_num_mol == 0 || _molecules == NULL) return;

       One_Card_Info_Format *pCard = _CardInfo[LINK_TOKEN-1].Head();
       while (pCard) pCard = _CardInfo[LINK_TOKEN-1].delete_a_card(pCard);
       output_link(_molecules[0]);

       if (_CardInfo[LINK_TOKEN-1].Head() == NULL) return;

       char buffer[300]; 
       int numRemark = 0;
       sprintf(buffer, "%s.link", File_ID);
       FILE *fp = fopen(buffer, "w");
       print_general_cards(fp, LINK_TOKEN-1, NDB_FILE_FORMAT_PDB, numRemark, 0);
       fprintf(fp, "\n");
       fclose (fp);
}

void _VALIDATION::write_sugar_links(const char *File_ID)
{
       if (_num_mol == 0 || _molecules == NULL) return;

       FILE *fp = NULL;
       char buffer[300];
       sprintf(buffer, "%s.sugar", File_ID);
       int count = 0;
       std::vector<_RESIDUE*> res_list;
       std::vector<std::vector<unsigned int> > connects;
       std::vector<std::vector<std::vector<unsigned int> > > linkages;
       std::set<unsigned int> index_set;

       _CHAIN *chain = _molecules[0]->GetFirstChain();
       while (chain) {
            if (chain->chain_type() != ATOMS_TOKEN || chain->ResidueNumbers() < 2) {
                 chain = _molecules[0]->GetNextChain();
                 continue;
            }

            res_list.clear();
            _RESIDUE *res = chain->GetFirstResidue();
            while (res) {
                 res_list.push_back(res);
                 res = chain->GetNextResidue();
            }
            if (res_list.size() < 2) {
                 chain = _molecules[0]->GetNextChain();
                 continue;
            }

            if (!get_connections(res_list, connects)) {
                 chain = _molecules[0]->GetNextChain();
                 continue;
            }

            get_linkages(connects, linkages);
            if (linkages.empty()) {
                 chain = _molecules[0]->GetNextChain();
                 continue;
            }

            for (unsigned int i = 0; i < linkages.size(); ++i) {
                 index_set.clear();
                 for (unsigned int j = 0; j < linkages[i].size(); ++j) {
                      for (unsigned int k = 0; k < linkages[i][j].size(); ++k) {
                           index_set.insert(linkages[i][j][k]);
                      }
                 }
                 if (index_set.empty()) continue;

                 if (fp == NULL) {
                      fp = fopen(buffer, "w");
                      fprintf(fp, "The following residues are linked together.\n");
                 }
                 count++;
                 fprintf(fp, "\n%d\n", count);
                 for (std::set<unsigned int>::iterator pos = index_set.begin();
                                                  pos != index_set.end(); ++pos) {
                      fprintf(fp, "%s %s %d\n", res_list[*pos]->ResName(),
                                  chain->PDB_ChainID(), res_list[*pos]->pdb_res_no());
                 }

                 write_sugar_linkages(fp, linkages[i], res_list);
            }
            
            chain = _molecules[0]->GetNextChain();
       }

       if (fp) fclose(fp);
}

void _VALIDATION::output_link(_MOLECULE *mol)
{
       char SymOP[10];
       for (int i = 0; i < _n_link; i++) {
            _CHAIN *ch1 = mol->GetIndexChain(_link[i]->FstIndex);
            _CHAIN *ch2 = mol->GetIndexChain(_link[i]->SndIndex);
            if (!ch1 || !ch2) continue;
            _FIELD *Init = ch1->SeqResIndex(_link[i]->FstRes);
            _FIELD *End  = ch2->SeqResIndex(_link[i]->SndRes);
            if (!Init || !End) continue;
            One_Card_Info_Format *pCard = _CardInfo[LINK_TOKEN-1].
                    add_a_new_card(NdbToken[LINK_TOKEN-1], 0);
            pCard->AddElementToField(_link[i]->FstAtom, 1);
            pCard->AddElementToField(_link[i]->altLoc1, 2);
            pCard->AddElementToField(_link[i]->FstName, 3);
            pCard->AddElementToField(ch1->PDB_ChainID(), 4);
            pCard->AddElementToField(Init->Field[4], 5);
            pCard->AddElementToField(Init->InsCode, 6);
            pCard->AddElementToField(_link[i]->SndAtom, 7);
            pCard->AddElementToField(_link[i]->altLoc2, 8);
            pCard->AddElementToField(_link[i]->SndName, 9);
            pCard->AddElementToField(ch2->PDB_ChainID(), 10);
            pCard->AddElementToField(End->Field[4], 11);
            pCard->AddElementToField(End->InsCode, 12);
            get_pdbformat_symmetry(SymOP, _link[i]->SymOP_1);
            pCard->AddElementToField(SymOP, 13);
            get_pdbformat_symmetry(SymOP, _link[i]->SymOP_2);
            pCard->AddElementToField(SymOP, 14);
            pCard->AddElementToField(_link[i]->dist.Text(), 15);
       }
}

void _VALIDATION::check_terminal_atoms(CifString &message)
{
       int first = 1, l = 0, j = 0, k = 0, end;
       char buffer[100];
       _CHAIN *chain = NULL;
       _RESIDUE *residue = NULL;
       _ATOM *atom = NULL;
       const char **terminal_atoms = NULL;

       message.Clear();
       for (l = 0; l < _num_mol; l++) {
            chain = _molecules[l]->GetFirstChain();
            while (chain) {
                 if (chain->chain_type() != ATOMP_TOKEN &&
                     chain->chain_type() != ATOMN_TOKEN) {
                      chain = _molecules[l]->GetNextChain();
                      continue;
                 }
                 for (j = 0; j < chain->SeqLen(); j++) {
                      if (chain->SeqRes(j)->Deleted) continue;
                      residue = chain->SeqRes(j)->Residue;
                      if (!residue) continue;
                      if (chain->chain_type() == ATOMP_TOKEN) {
                           int start = 0;
                           int is_linked = 0;
                           if (!strcmp(residue->ResName(), "ASP") ||
                               !strcmp(residue->ResName(), "GLU"))
                                is_linked = is_residue_linked(residue);
                           if (j == 0) {
                                terminal_atoms = _AA_C_Terminal_Atoms;
                                if (is_linked == 2) start = 1;
                                end = NUM_AA_C_TERMINAL_ATOMS;
                           } else if (j == (chain->SeqLen() - 1)) {
                                terminal_atoms = _AA_N_Terminal_Atoms;
                                end = NUM_AA_N_TERMINAL_ATOMS;
                           } else {
                                terminal_atoms = _AA_Terminal_Atoms;
                                if (is_linked == 2) start = 1;
                                end = NUM_AA_TERMINAL_ATOMS;
                           }
                           for (k = start; k < end; k++) {
                                atom = residue->find_atom(terminal_atoms[k]);
                                if (is_terminal_atom(atom, residue, terminal_atoms[k])) {
                                     if (first) {
                                          message += "\n    CHECK TERMINAL ATOMS\n";
                                          message += "    --------------------\n\n";
                                          message += "Terminal atom(s) showed in";
                                          message += " middle of sequence will be deleted:\n\n";
                                          first = 0;
                                     }
                                     sprintf(buffer, "%s %s %d%s %s\n", atom->atmtype(),
                                             atom->pdb_resnam(), atom->pdb_resnum(),
                                             atom->ins_code(), atom->pdb_chnid());
                                     message += buffer;
                                     _Error_Check[3]++;
                                }
                           }
                      } else if (chain->chain_type() == ATOMN_TOKEN) {
                           if (j == 0) {
                                terminal_atoms = _NA_3_Terminal_Atoms;
                                end = NUM_NA_3_TERMINAL_ATOMS;
                           } else if (j == (chain->SeqLen() - 1)) {
                                terminal_atoms = _NA_5_Terminal_Atoms;
                                end = NUM_NA_5_TERMINAL_ATOMS;
                           } else {
                                terminal_atoms = _NA_Terminal_Atoms;
                                end = NUM_NA_TERMINAL_ATOMS;
                           } 
                           for (k = 0; k < end; k++) {
                                atom = residue->find_atom(terminal_atoms[k]);
                                if (is_terminal_atom(atom, residue, terminal_atoms[k])) {
                                     if (first) {
                                          message += "\n    CHECK TERMINAL ATOMS\n";
                                          message += "    --------------------\n\n";
                                          message += "Terminal atom(s) showed in";
                                          message += " middle of sequence will be deleted:\n\n";
                                          first = 0;
                                     }
                                     sprintf(buffer, "%s %s %d%s %s\n", atom->atmtype(),
                                             atom->pdb_resnam(), atom->pdb_resnum(),
                                             atom->ins_code(), atom->pdb_chnid());
                                     message += buffer;
                                     _Error_Check[3]++;
                                }
                           }
                      }
                 }
                 chain = _molecules[l]->GetNextChain();
            }
       }
}

void _VALIDATION::pdbx_find_outside_torsions(FILE *fp)
{
       int first = 1;
       char buffer[100], chain_id[5], ins_code[5];
/*
       CifString Header;
       Header.Copy("\nRamachandran outliers:\n");
       if (_num_mol > 1)
            Header += "      Residue  Chain  Sequence  Model  ";
       else Header += "      Residue  Chain  Sequence         ";
       Header += "     PSI       PHI\n";
*/
       _n_t_pair_outliers = 0;
       _n_t_pair_checked = 0;
       for (int i = 0; i < _n_t_pairs; i++) {
            if ((strcmp(_t_pairs[i]->ResName, "GLY") && is_L_aa_residue(_t_pairs[i]->ResName)) ||
                is_D_aa_residue(_t_pairs[i]->ResName)) _n_t_pair_checked++;

            if ((strcmp(_t_pairs[i]->ResName, "GLY") &&
                is_L_aa_residue(_t_pairs[i]->ResName) &&
/*
                strcmp(_t_pairs[i]->ResName, "SAR") &&
                strcmp(_t_pairs[i]->ResName, "PRO") &&
*/
                is_torsion_out_of_regions1(_t_pairs[i]->val, _t_pairs[i]->ept)) ||
               (is_D_aa_residue(_t_pairs[i]->ResName) &&
                is_torsion_out_of_regions1(-_t_pairs[i]->val, -_t_pairs[i]->ept))) {
                 _n_t_pair_outliers++;
                 if (first) {
                      fprintf(fp, "\n==> The following table contains a list of torsion angles outside the expected\n");
                      fprintf(fp, "    Ramachandran regions [GJ. Kleywegt and TA. Jones, PHI/PSI-chology:\n");
                      fprintf(fp, "    Ramachandran Revisited, Structure  1996, 4, 1395 - 1400].\n\n");
                      if (_num_mol > 1)
                           fprintf(fp, "      Residue  Chain  Sequence  Model  ");
                      else fprintf(fp, "      Residue  Chain  Sequence         ");
                      fprintf(fp, "     PSI       PHI\n");
                 }
                 first = 0;
                 strcpy(chain_id, " ");
                 if (strcmp(_t_pairs[i]->ChainID, "")) strcpy(chain_id, _t_pairs[i]->ChainID);
                 strcpy(ins_code, " ");
                 if (strcmp(_t_pairs[i]->ins_code, "")) strcpy(ins_code, _t_pairs[i]->ins_code);
                 if (_num_mol > 1)
                      sprintf(buffer, "        %3s      %c     %4d%c    %3d     %7.2f   %7.2f\n",
                              _t_pairs[i]->ResName, chain_id[0], _t_pairs[i]->ResNum, ins_code[0],
                              _t_pairs[i]->Mol_ID, _t_pairs[i]->ept, _t_pairs[i]->val);
                 else sprintf(buffer, "        %3s      %c     %4d%c            %7.2f   %7.2f\n",
                              _t_pairs[i]->ResName, chain_id[0], _t_pairs[i]->ResNum, ins_code[0],
                              _t_pairs[i]->ept, _t_pairs[i]->val);
                 fprintf(fp, "%s", buffer);
/*
                 if (_generate_flag) {
                      if (Header != "") {
                           _bond_angle_torsion += Header;
                           Header.Clear();
                      }
                      _bond_angle_torsion += buffer;
                 }
*/
            }
       }
}

void _VALIDATION::pdbx_find_non_cis_trans_torsions(FILE *fp)
{
       if (_n_torsion == 0 || _torsion == NULL) return;
       if (_torsionIndex.Root() < 0) return;

       int first = 1;
       char chain_id1[5], ins_code1[5], chain_id2[5], ins_code2[5];
       CifString index;
       index.Copy("OMEGA");

       int ret = _torsionIndex.Seek(index);
       while (ret >= 0) {
            if (index != _torsionIndex.GetCurrData()) break;
            if (fabs(_torsion[ret]->val) < 30.0 ||
                fabs(_torsion[ret]->val - 180.0) < 30.0 ||
                fabs(_torsion[ret]->val + 180.0) < 30.0) {
                 ret = _torsionIndex.GoToNext();
                 continue;
            }
            if (first) {
                 fprintf(fp, "\n==> The following peptide bonds deviate significantly from both cis\n");
                 fprintf(fp, "    and trans conformation. Trans is defined as 180 +/- 30 and cis\n");
                 fprintf(fp, "    is defined as 0 +/- 30 degrees.\n\n");
                 if (_num_mol > 1)
                      fprintf(fp, "  Residue  Chain  Sequence    Residue  Chain  Sequence Model  ");
                 else fprintf(fp, "  Residue  Chain  Sequence    Residue  Chain  Sequence        ");
                 fprintf(fp, "     OMEGA\n");
            }
            first = 0;
            strcpy(chain_id1, " ");
            if (strcmp(_torsion[ret]->fst->pdb_chnid(), ""))
                 strcpy(chain_id1, _torsion[ret]->fst->pdb_chnid());
            strcpy(ins_code1, " ");
            if (strcmp(_torsion[ret]->fst->ins_code(), ""))
                 strcpy(ins_code1, _torsion[ret]->fst->ins_code());
            strcpy(chain_id2, " ");
            if (strcmp(_torsion[ret]->fth->pdb_chnid(), ""))
                 strcpy(chain_id2, _torsion[ret]->fth->pdb_chnid());
            strcpy(ins_code2, " ");
            if (strcmp(_torsion[ret]->fth->ins_code(), ""))
                 strcpy(ins_code2, _torsion[ret]->fth->ins_code());
            if (_num_mol > 1)
                 fprintf(fp, "    %3s      %c     %4d%c        %3s      %c     %4d%c    %3d      %7.2f\n",
                          _torsion[ret]->fst->pdb_resnam(), chain_id1[0],
                          _torsion[ret]->fst->pdb_resnum(), ins_code1[0],
                          _torsion[ret]->fth->pdb_resnam(), chain_id2[0],
                          _torsion[ret]->fth->pdb_resnum(), ins_code2[0],
                          _torsion[ret]->Mol_ID, _torsion[ret]->val);
            else fprintf(fp, "    %3s      %c     %4d%c        %3s      %c     %4d%c             %7.2f\n",
                          _torsion[ret]->fst->pdb_resnam(), chain_id1[0],
                          _torsion[ret]->fst->pdb_resnum(), ins_code1[0],
                          _torsion[ret]->fth->pdb_resnam(), chain_id2[0],
                          _torsion[ret]->fth->pdb_resnum(), ins_code2[0],
                          _torsion[ret]->val);
            ret = _torsionIndex.GoToNext();
       }
}

void _VALIDATION::pdbx_find_main_chain_planarity(FILE *fp)
{
       if (_n_torsion == 0 || _torsion == NULL) return;
       if (_torsionIndex.Root() < 0) return;

       int first = 1;
       char chain_id[5], ins_code[5], buffer[200];
       CifString index;
       index.Copy("PLANAR");
       int ret = _torsionIndex.Seek(index);
       while (ret >= 0) {
            if (index != _torsionIndex.GetCurrData()) break;
            if (fabs(_torsion[ret]->val) <= MAINPLANEOFF) {
                 ret = _torsionIndex.GoToNext();
                 continue;
            }
            if (first) {
                 fprintf(fp, "\n==> The following residues have a pseudo planarity torsion,\n");
                 fprintf(fp, "    C(i) - Ca(i) - N(i+1) - O(i), greater 10.0 degrees.\n\n");
                 if (_num_mol > 1)
                      fprintf(fp, "      Residue  Chain  Sequence  Model  ");
                 else fprintf(fp, "      Residue  Chain  Sequence         ");
                 fprintf(fp, "   ANGLE\n");
            }
            first = 0;
            strcpy(chain_id, " ");
            if (strcmp(_torsion[ret]->fst->pdb_chnid(), ""))
                 strcpy(chain_id, _torsion[ret]->fst->pdb_chnid());
            strcpy(ins_code, " ");
            if (strcmp(_torsion[ret]->fst->ins_code(), ""))
                 strcpy(ins_code, _torsion[ret]->fst->ins_code());

            if (_num_mol > 1)
                 fprintf(fp, "        %3s      %c     %4d%c    %3d     %7.2f\n",
                          _torsion[ret]->fst->pdb_resnam(), chain_id[0],
                          _torsion[ret]->fst->pdb_resnum(), ins_code[0],
                          _torsion[ret]->Mol_ID, _torsion[ret]->val);
            else fprintf(fp, "        %3s      %c     %4d%c            %7.2f\n",
                          _torsion[ret]->fst->pdb_resnam(), chain_id[0],
                          _torsion[ret]->fst->pdb_resnum(), ins_code[0],
                          _torsion[ret]->val);
            ret = _torsionIndex.GoToNext();
       }

       CifString cs, Header;
       first = 1;
       if (_num_mol > 1)
            Header.Copy("      Residue  Chain  Sequence  Model  ");
       else Header.Copy("      Residue  Chain  Sequence         ");
       Header += "   Improper  Details\n";
       const char *_tor_name[3] = { "CALPHA", "C1'NA", "CBETA" };
  for (int i = 0; i < 3; ++i) {
       index.Copy(_tor_name[i]);
       ret = _torsionIndex.Seek(index);
       while (ret >= 0) {
            if (index != _torsionIndex.GetCurrData()) break;
            check_chiral_error(cs, _torsion[ret]);
            if (cs == "") {
                 ret = _torsionIndex.GoToNext();
                 continue;
            }
            if (first) {
                 fprintf(fp, "\n==> The following residues have unexpected configuration of the \n");
                 fprintf(fp, "    chiral center using CA--C--CB--N chirality for amino acids\n");
                 fprintf(fp, "    and C1'--O4'--N1(N9)--C2' chirality for nucleic acids.\n\n");
                 if (_num_mol > 1)
                      fprintf(fp, "      Residue  Chain  Sequence  Model  ");
                 else fprintf(fp, "      Residue  Chain  Sequence         ");
                 fprintf(fp, "   Improper  Details\n");
            }
            first = 0;
            strcpy(chain_id, " ");
            if (strcmp(_torsion[ret]->fst->pdb_chnid(), ""))
                 strcpy(chain_id, _torsion[ret]->fst->pdb_chnid());
            strcpy(ins_code, " ");
            if (strcmp(_torsion[ret]->fst->ins_code(), ""))
                 strcpy(ins_code, _torsion[ret]->fst->ins_code());

            if (_num_mol > 1)
                 sprintf(buffer, "        %3s      %c     %4d%c    %3d       %7.2f   %s\n",
                          _torsion[ret]->fst->pdb_resnam(), chain_id[0],
                          _torsion[ret]->fst->pdb_resnum(), ins_code[0],
                          _torsion[ret]->Mol_ID, _torsion[ret]->val, cs.Text());
            else sprintf(buffer, "        %3s      %c     %4d%c              %7.2f   %s\n",
                          _torsion[ret]->fst->pdb_resnam(), chain_id[0],
                          _torsion[ret]->fst->pdb_resnum(), ins_code[0],
                          _torsion[ret]->val, cs.Text());
            fprintf(fp, "%s", buffer);
            if (_generate_flag && 
               (strstr(buffer, CHIRAL_CUTOFF) ||
                strstr(buffer, CHIRAL_CUTOFF1) ||
                strstr(buffer, CHIRAL_CUTOFF2))) {
                 if (Header != "") {
                      _chirality += Header;
                      Header.Clear();
                 }
                 _chirality += buffer; 
            }
            ret = _torsionIndex.GoToNext();
       }
  }
}

void _VALIDATION::pdbx_find_side_chain_plane(FILE *fp)
{
       if (!_molecules) return;

       ISTable *Table = get_plane_outlier_table(_num_mol, _molecules);
       if (!Table->GetNumRows()) {
            delete Table; return;
       }

       int rowNo = Table->GetNumRows();
       int model_id = Table->GetColumnIndex("PDB_model_num");
       int auth_asym_id = Table->GetColumnIndex("auth_asym_id");
       int auth_comp_id = Table->GetColumnIndex("auth_comp_id");
       int auth_seq_id = Table->GetColumnIndex("auth_seq_id");
       int insert_id = Table->GetColumnIndex("PDB_ins_code");
       int rmsd_index = Table->GetColumnIndex("rmsd");

       CifString chnid, resnam, inscode, cifstring;
       int resnum, Mol_ID;
       float rmsd;

       fprintf(fp, "\n==> Side chain planar groups in the following residues have a total\n");
       fprintf(fp, "    RMS distance of all atoms from the best-fit plane by more than\n");
       fprintf(fp, "    an expected value of 6*RMSD, with an RMSD 0.02 angstroms, or at\n");
       fprintf(fp, "    least one atom has an RMSD greater than this value.\n\n");
       if (_num_mol > 1)
            fprintf(fp, "      Residue  Chain  Sequence  Model  ");
       else fprintf(fp, "      Residue  Chain  Sequence         ");
       fprintf(fp, "     RMS\n");

       for (int i = 0; i < rowNo; i++) {
            get_value(chnid, Table, auth_asym_id, i);
            get_value(resnam, Table, auth_comp_id, i);
            get_value(cifstring, Table, auth_seq_id, i);
            resnum = atoi(cifstring.Text());
            get_value(inscode, Table, insert_id, i);
            if (chnid == "") chnid.Copy(" ");
            if (inscode == "") inscode.Copy(" ");

            get_value(cifstring, Table, rmsd_index, i);
            rmsd = atof(cifstring.Text());
            get_value(cifstring, Table, model_id, i);
            Mol_ID = atoi(cifstring.Text());

            if (_num_mol > 1)
                 fprintf(fp, "        %3s      %c     %4d%c    %3d     %7.2f\n",
                          resnam.Text(), chnid.Text()[0], resnum,
                          inscode.Text()[0], Mol_ID, rmsd);
            else fprintf(fp, "        %3s      %c     %4d%c            %7.2f\n",
                          resnam.Text(), chnid.Text()[0], resnum,
                          inscode.Text()[0], rmsd);

       }
       delete Table;
}

static int find_letter_token(char *line, char* token)
{
       int i ,len;
       if (line == NULL || token == NULL) return FALSE;

       strcpy(token, "");
       if (strncmp(line, "$$@", 3)) return FALSE;
       else {
            len = strlen(line);
            for (i = 4; i < len; i++)
                 if (!strncmp("@$$", &line[i], 3)) break;
            if (i == len) return FALSE;
            else {
                 strncpy(token, &line[3], i-3);
                 token[i-3] = 0;
                 return (i + 2);
            }
       }
}

void ndb_writing_missing_words(FILE *infile, FILE *outfile)
{
       char line[MxLen], token[MxStrLen];
       int ret = FALSE, i, len;
       do {
            if (feof(infile)) break;
            if (fgets(line, MxLen, infile));
            len = strlen(line);
            for (i = 0; i < len; i++)  {
                 ret = find_letter_token(&line[i], token);
                 if (ret == FALSE) fprintf(outfile, "%c", line[i]);
                 else break;
            }
       } while (ret == FALSE);
}

int skip_contact(CONTACT *contact)
{
       if ((contact->a_occupancy < 1.0 && contact->b_occupancy < 1.0) ||
           (contact->a_occupancy < 1.0 && (!strcmp(contact->a_resnam, "HOH") || !strcmp(contact->a_resnam, "DOD"))) ||
           (contact->b_occupancy < 1.0 && (!strcmp(contact->b_resnam, "HOH") || !strcmp(contact->b_resnam, "DOD"))))
            return 1;
       return 0;
}

int is_L_aa_residue(const char* ResName)
{
       ConnectFormat *drug = find_drug(ResName);
       if (drug && !strcasecmp(drug->type(), "L-peptide linking"))
            return 1;
       return 0;
}

int is_D_aa_residue(const char* ResName)
{
       ConnectFormat *drug = find_drug(ResName);
       if (drug && !strcasecmp(drug->type(), "D-peptide linking"))
            return 1;
       return 0;
}

void init_cbeta_library()
{
       _cbeta_index.clear();
       for (int i = 0; i < NUM_CBETA_CHIRAL; ++i) {
            _cbeta_index.insert(std::make_pair(_cbeta_chiral[i].res, i));
       }
}

void check_chiral_error(CifString &cs, _TORSION *torsion)
{
       cs.Clear();

       ConnectFormat *drug = find_drug(torsion->fst->pdb_resnam());
       if (drug) {
            _SIndex *index = drug->find_atom(torsion->fst->pdb_atmnam());
            if (index) {
                 if (drug->atoms(index->pos())->stereo_config() &&
                    (!strcmp(drug->atoms(index->pos())->stereo_config(), "") ||
                     !strcmp(drug->atoms(index->pos())->stereo_config(), "N"))) return;
            }
       }
       if (!strcmp(torsion->Rep_Name, "CALPHA"))
            check_calpha_chiral_error(cs, torsion);
       else if (!strcmp(torsion->Rep_Name, "CBETA"))
            check_cbeta_chiral_error(cs, torsion);
       else if (!strcmp(torsion->Rep_Name, "C1'NA"))
            check_c1na_chiral_error(cs, torsion);
}

void get_pdbformat_symmetry(char *symm, const char *old_symm)
{
       *symm = '\0';
       if (!strcmp(old_symm, "")) strcpy(symm, "1555");
       else {
           for (unsigned int i = 0; i < strlen(old_symm); i++) {
                if (isdigit(old_symm[i])) {
                     *symm = old_symm[i];
                     symm++;
                }
           }
           *symm = '\0';
       }
}

float cal_min_distance(_MOLECULE *mol, _ATOM *atom, const double& cutoff)
{
       float min_distance = 100.0;
       double dist = 0;

       _RESIDUE *Residue = NULL;
       _ATOM *atom1 = NULL;
       _CHAIN *chain = mol->GetFirstChain();
       while (chain) {
            if (chain->op() != 1) {
                 chain = mol->GetNextChain();
                 continue;
            }
            Residue = chain->GetFirstResidue();
            if (chain->chain_type() == HETAS_TOKEN &&
               (!strcmp(Residue->ResName(), "HOH") ||
                !strcmp(Residue->ResName(), "DOD"))) {
                 chain = mol->GetNextChain();
                 continue;
            }
            while (Residue) {
                 atom1 = Residue->GetFirstAtom();
                 while (atom1) {
                      dist = cal_dist(atom, atom1, cutoff);
                      if (dist < min_distance) min_distance = dist;
                      atom1 = Residue->GetNextAtom();
                 }
                 Residue = chain->GetNextResidue();
            }
            chain = mol->GetNextChain();
       }
       return (min_distance);
}

static void check_calpha_chiral_error(CifString &cs, _TORSION *torsion)
{
       int is_standard = 0;
       // if (is_standard_aa_residue(torsion->fst->pdb_resnam())) is_standard = 1;
       if (is_aa_residue(torsion->fst->pdb_resnam())) is_standard = 1;

       CifString type;
       type.Clear();
/*
       if (is_standard)
            type.Copy("L");
       else */ if (/* !strcmp(torsion->fst->pdb_resnam(), "DGN") || */
                !strcmp(torsion->fst->pdb_resnam(), "DNE") /* ||
                !strcmp(torsion->fst->pdb_resnam(), "MHO") */ )
            type.Copy("P");
       else {
            ConnectFormat *drug = find_drug(torsion->fst->pdb_resnam());
            if (drug) {
                 if (!strcasecmp(drug->type(), "L-peptide linking"))
                      type.Copy("L");
                 else if (!strcasecmp(drug->type(), "D-peptide linking"))
                      type.Copy("D");
            }
       }
       if (type == "") return;

       double Dvalue = -35.0;
       double Lvalue = 35.0;
       double DLerror = 10.0;
       if (type == "P") {
            if (fabs(torsion->val) <= 170.0 ||
                fabs(torsion->val) >= 10.0) {
                 cs.Copy("EXPECTING PLANAR");
            }
       } else if (torsion->val < 0.0) {
            if (type == "L") {
                 // if (torsion->val > (Dvalue - DLerror) &&
                 //    torsion->val < (Dvalue + DLerror))
                      cs.Copy("Expecting L Found D WRONG HAND");
                 // else cs.Copy("Expecting L Found D OUTSIDE RANGE");
            } else if (type == "D") {
                 if (torsion->val > (Dvalue - DLerror) &&
                     torsion->val < (Dvalue + DLerror)) {}
                 else cs.Copy("Expecting D Found D OUTSIDE RANGE");
            }
       } else if (torsion->val > 0.0) {
            if (type == "L") {
                 if (torsion->val > (Lvalue - DLerror) &&
                     torsion->val < (Lvalue + DLerror)) {}
                 else cs.Copy("Expecting L Found L OUTSIDE RANGE");
            } else if (type == "D") {
                 // if (torsion->val > (Lvalue - DLerror) &&
                 //    torsion->val < (Lvalue + DLerror))
                      cs.Copy("Expecting D Found L WRONG HAND");
                 // else cs.Copy("Expecting D Found L OUTSIDE RANGE");
            }
       }
       if (is_standard && (fabs(torsion->val) > 170.0 || fabs(torsion->val) < 10.0)) {
            if (type == "L") {
                 if (torsion->val > 0)
                      cs.Copy("Expecting L Found L EXPECTING SP3");
                 else cs.Copy("Expecting L Found D EXPECTING SP3");
            } else if (type == "D") {
                 if (torsion->val > 0)
                      cs.Copy("Expecting D Found L EXPECTING SP3");
                 else cs.Copy("Expecting D Found D EXPECTING SP3");
            }
       }
}

static void check_cbeta_chiral_error(CifString &cs, _TORSION *torsion)
{
       std::map<std::string, int>::iterator pos =
                   _cbeta_index.find(torsion->fst->pdb_resnam());
       if (pos == _cbeta_index.end()) return;

       std::string cbc = _cbeta_chiral[pos->second].r_s;
       std::string cbti = _cbeta_chiral[pos->second].t_i;

       if (torsion->val < 0) {
            if ((cbc == "R" && cbti == "T") ||
                (cbc == "S" && cbti == "I")) {
                 if (cbti == "T") cs.Copy("Expecting R Found S CBETA WRONG HAND");
                 if (cbti == "I") cs.Copy("Expecting S Found R CBETA WRONG HAND");
            }
       } else if (torsion->val > 0) {
            if ((cbc == "R" && cbti == "I") ||
                (cbc == "S" && cbti == "T")) {
                 if (cbti == "T") cs.Copy("Expecting S Found R CBETA WRONG HAND");
                 if (cbti == "I") cs.Copy("Expecting R Found S CBETA WRONG HAND");
            }
       }
}

static void check_c1na_chiral_error(CifString &cs, _TORSION *torsion)
{
       ConnectFormat *drug = find_drug(torsion->fst->pdb_resnam());
       if (!drug) return;
       if (strcasecmp(drug->type(), "DNA linking") &&
           strcasecmp(drug->type(), "L-DNA linking") &&
           strcasecmp(drug->type(), "RNA linking") &&
           strcasecmp(drug->type(), "L-RNA linking") &&
           strcasecmp(drug->type(), "DNA OH 5 prime terminus") &&
           strcasecmp(drug->type(), "RNA OH 5 prime terminus") &&
           strcasecmp(drug->type(), "DNA OH 3 prime terminus") &&
           strcasecmp(drug->type(), "RNA OH 3 prime terminus")) return;
        
       torsion->val = mod(torsion->val, -180, 180);
       if (!strcasecmp(drug->type(), "L-DNA linking") || !strcasecmp(drug->type(), "L-RNA linking")) {
            if ((torsion->val > 45.0 && torsion->val < 60.0) ||
                (torsion->val > 10.0 && torsion->val < 25.0))
                 cs.Copy("Expecting L Found L OUTSIDE RANGE");

            if (torsion->val < 10.0 || torsion->val > 60.0)
                 cs.Copy("Expecting L Found D WRONG HAND"); 
       } else {
            if ((torsion->val < -45.0 && torsion->val > -60.0) ||
                (torsion->val < -10.0 && torsion->val > -25.0))
                 cs.Copy("Expecting D Found D OUTSIDE RANGE");

            if (torsion->val < -60.0 || torsion->val > -10.0)
                 cs.Copy("Expecting D Found L WRONG HAND"); 
       }
}

static void passing_missing_words(const int missing, FILE *infile, const char *oldtoken)
{
       char line[MxLen], token[MxStrLen];
       int i, ret = FALSE, len;
     
       if (missing == FALSE) {
            do {
                 if (feof(infile)) break;
                 if (fgets(line,MxLen,infile));
                 len = strlen(line);
                 for (i = 0; i < len; i++)  {
                      ret = find_letter_token(&line[i], token);
                      if (ret != FALSE && !strcmp(token, oldtoken)) 
                           break;
                 }
            } while (ret == FALSE);
       }
}

static int is_terminal_atom(_ATOM *atom, _RESIDUE *residue, const char *atom_name)
{
       if (atom == NULL) return 0;

       if (!strcmp(atom_name, "HO5'")) return 1;

       ConnectFormat *drug = find_drug(residue->ResName());
       if (drug == NULL) return 0;

       _SIndex *exist = drug->find_atom(atom_name);
       if (exist && !strcasecmp(drug->atoms(exist->pos())->leaving_atom_flag(), "y"))
           return 1;

       return 0;
}

static float cal_dist(_ATOM *atom_a, _ATOM *atom_b, const double& cutoff)
{
       float min_distance = 100.0;
/*
       if (!atom_a || !atom_b || strcmp(atom_a->alt_loc(), "") &&
           strcmp(atom_b->alt_loc(), "") &&
           strcmp(atom_a->alt_loc(), atom_b->alt_loc()))
            return min_distance;
*/
       if (!strcmp(atom_b->atom_type(), "N") ||
           !strcmp(atom_b->atom_type(), "O")) {
            float dist = (atom_a->orig().x - atom_b->orig().x)
                        * (atom_a->orig().x - atom_b->orig().x)
                        + (atom_a->orig().y - atom_b->orig().y)
                        * (atom_a->orig().y - atom_b->orig().y)
                        + (atom_a->orig().z - atom_b->orig().z)
                        * (atom_a->orig().z - atom_b->orig().z);
            dist = sqrt(dist);
            if (dist > (cutoff - 0.001)) {
                 if (dist < min_distance) min_distance = dist;
            }
       }
       return min_distance;
}

static bool get_connections(const std::vector<_RESIDUE*>& res_list,
                            std::vector<std::vector<unsigned int> >& connects)
{
       connects.clear();

       std::vector<unsigned int> index;
       index.clear();
       for (unsigned int i = 0; i < res_list.size(); ++i) {
            connects.push_back(index);
       }

       bool found = false;
       for (unsigned int i = 0; i < res_list.size() - 1; ++i) {
            for (unsigned int j = i + 1; j < res_list.size(); ++j) {
                 if (is_connect(res_list[i], res_list[j], "O", "C")) {
                      connects[i].push_back(j);
                      connects[j].push_back(i);
                      found = true;
                 }
            }
       }
       return found;
}

static void write_sugar_linkages(FILE *fp, const std::vector<std::vector<unsigned int> >&
                                 linkages, const std::vector<_RESIDUE*>& res_list)
{
       for (unsigned int j = 0; j < linkages.size(); ++j) {
            int max_column = 12;
            if (linkages.size() > 1) max_column = 11;
            int l = linkages[j].size() / max_column;
            int x = linkages[j].size() % max_column;
            int m = l;
            if (x == 0) m--;
            for (int i = 0; i <= m; ++i) {
                 fprintf(fp, "\n");
                 int n = max_column;
                 if (i == l) n = x;
                 if (i == 0) {
                      if (linkages.size() > 1) {
                           if (j == 0)
                                fprintf(fp, "Main:   ");
                           else fprintf(fp, "Branch: ");
                      }
                 } else {
                      if (linkages.size() > 1) {
                           fprintf(fp, "        ");
                      }
                 }
                 for (int y = 0; y < n; ++y) {
                      int k = i * max_column + y;
                      if (y) fprintf(fp, " -- ");
                      fprintf(fp, "%-3s", res_list[linkages[j][k]]->ResName());
                 }
                 fprintf(fp, "\n");

                 if (linkages.size() > 1) fprintf(fp, "        ");
                 for (int y = 0; y < n; ++y) {
                      int k = i * max_column + y;
                      if (y) fprintf(fp, "   ");
                      fprintf(fp, "%-3d", res_list[linkages[j][k]]->pdb_res_no());
                 }
                 fprintf(fp, "\n");
            }
       }
}

static void get_linkages(const std::vector<std::vector<unsigned int> >& connects,
                         std::vector<std::vector<std::vector<unsigned int> > >& linkages)
{
       linkages.clear();

       std::vector<std::set<unsigned int> > groups;

       clustering(connects, groups);

       std::map<unsigned int, unsigned int> forward_mapping, reverse_mapping;
       std::vector<std::vector<std::pair<unsigned int, bool> > > local_connects;
       std::vector<std::pair<unsigned int, bool> > data;
       std::vector<std::vector<unsigned int> > links;

       for (unsigned int i = 0; i < groups.size(); ++i) {
            if (groups[i].size() < 2) continue;

            forward_mapping.clear();
            reverse_mapping.clear();
            local_connects.clear();
            data.clear();
            for (std::set<unsigned int>::iterator spos = groups[i].begin();
                                             spos != groups[i].end(); ++spos) {
                 forward_mapping.insert(std::make_pair(*spos, local_connects.size()));
                 reverse_mapping.insert(std::make_pair(local_connects.size(), *spos));
                 local_connects.push_back(data);
            }

            bool found = false;
            for (unsigned int j = 0; j < local_connects.size(); ++j) {
                 std::map<unsigned int, unsigned int>::iterator mpos1
                                 = reverse_mapping.find(j);
                 if (mpos1 == reverse_mapping.end()) continue;
                 if (mpos1->second >= connects.size()) continue;
                 for (unsigned int k = 0; k < connects[mpos1->second].size(); ++k) {
                      std::map<unsigned int, unsigned int>::iterator mpos2 =
                               forward_mapping.find(connects[mpos1->second][k]);
                      if (mpos2 == forward_mapping.end()) continue;
                      local_connects[j].push_back(std::make_pair(mpos2->second, false));
                      local_connects[mpos2->second].push_back(std::make_pair(j, false));
                      found = true;
                 }
            }
            if (!found) continue;

            find_linkages(local_connects, links);
            if (links.empty()) continue;

            found = false;
            for (unsigned int j = 0; j < links.size(); ++j) {
                 for (unsigned int k = 0; k < links[j].size(); ++k) {
                      std::map<unsigned int, unsigned int>::iterator pos =
                              reverse_mapping.find(links[j][k]);
                      if (pos == reverse_mapping.end()) {
                           found = true;
                           continue;
                      }
                      links[j][k] = pos->second;
                 }
            }
            if (!found) linkages.push_back(links);
       }
}

static void clustering(const std::vector<std::vector<unsigned int> >& connects,
                       std::vector<std::set<unsigned int> >& groups)
{
       groups.clear();

       for (unsigned int i = 0; i < connects.size(); ++i) {
            for (unsigned int j = 0; j < connects[i].size(); ++j) {
                 int ipos = -1;
                 int jpos = -1;
                 for (unsigned int k = 0; k < groups.size(); ++k) {
                      if (groups[k].find(i) != groups[k].end()) ipos = k;
                      if (groups[k].find(connects[i][j]) != groups[k].end()) jpos = k;
                 }
                 if (ipos >= 0 && jpos >= 0) {
                      if (ipos != jpos) {
                           for (std::set<unsigned int>::iterator pos = groups[jpos].begin();
                                                    pos != groups[jpos].end(); ++pos) {
                                groups[ipos].insert(*pos);
                           }
                           groups[jpos].clear();
                      }
                      groups[ipos].insert(i);
                      groups[ipos].insert(connects[i][j]);
                 } else if (ipos >= 0) {
                      groups[ipos].insert(connects[i][j]);
                 } else if (jpos >= 0) {
                      groups[jpos].insert(i);
                 } else {
                      std::set<unsigned int> index_set;
                      index_set.clear();
                      index_set.insert(i);
                      index_set.insert(connects[i][j]);
                      groups.push_back(index_set);
                 }
            }
       }
}

static void find_linkages(std::vector<std::vector<std::pair<unsigned int, bool> > >&
                          connects, std::vector<std::vector<unsigned int> >& linkages)
{
       linkages.clear();

       std::vector<unsigned int> stack, longest_path;
       longest_path.clear();

       std::vector<bool> labels;
       labels.clear();
       for (unsigned int i = 0; i < connects.size(); ++i) labels.push_back(false);  

       for (unsigned int i = 0; i < connects.size(); ++i) {
            // clear previous node's search path
            for (unsigned int j = 0; j < connects.size(); ++j) {
                 labels[j] = false;
                 for (unsigned k = 0; k < connects[j].size(); ++k) {
                      connects[j][k].second = false;
                 }
            }

            stack.clear();
            labels[i] = true;
            stack.push_back(i);

            bool found_link = false;
            do {
                 while (true) {
                      bool found = false;
                      if (!stack.empty()) {
                           int j = stack[stack.size() - 1];
                           for (unsigned k = 0; k < connects[j].size(); ++k) {
                                if (connects[j][k].second) continue;
                                connects[j][k].second = true;
                                if (labels[connects[j][k].first]) continue;
                                labels[connects[j][k].first] = true;
                                stack.push_back(connects[j][k].first);
                                found = true;
                                found_link = true;
                                break;
                           }
                      }
                      if (!found) break;
                 }

                 if (found_link) {
                      found_link = false;
                      if (stack.size() > longest_path.size()) longest_path = stack;
                 }

                 if (!stack.empty()) {
                      int j = stack[stack.size() - 1];
                      stack.pop_back();
                      labels[j] = false;
                      for (unsigned k = 0; k < connects[j].size(); ++k) {
                           connects[j][k].second = false;
                      }
                 }
            } while (!stack.empty());
       }

       if (longest_path.empty()) return;

       if (longest_path.size() == connects.size()) {
            linkages.push_back(longest_path);
            return;
       }
}
